import numpy as np
import pandas as pd
from xarray import Dataset

from foxes.utils import wd2uv, write_nc
import foxes.variables as FV
import foxes.constants as FC


def calc_point_results(
    algo,
    g_pts,
    farm_results=None,
    seq_iter=None,
    runner=None,
    verbosity=0,
    **kwargs,
):
    """
    Helper function that calculates results at grid points.

    Parameters
    ----------
    algo: foxes.Algorithm
        The algorithm for point calculation
    g_pts: numpy.ndarray
        The grid points, shape: (n_states, n_x, n_y, 3)
    farm_results: xarray.Dataset, optional
        The farm results
    seq_iter: foxes.algorithms.sequential.SequentialIter, optional
        The sequential itarator
    runner: foxes.utils.runners.Runner, optional
        The runner
    verbosity: int
        The verbosity level, 0 = silent
    kwargs: dict, optional
        Additional parameters for algo.calc_points

    """
    averb = None if verbosity == algo.verbosity else algo.verbosity
    if averb is not None:
        algo.verbosity = verbosity
    fres = farm_results if seq_iter is None else seq_iter.farm_results
    if runner is None:
        point_results = algo.calc_points(fres, points=g_pts, **kwargs)
    else:
        kwargs["points"] = g_pts
        point_results = runner.run(algo.calc_points, args=(fres,), kwargs=kwargs)
    if averb is not None:
        algo.verbosity = averb

    return point_results


def get_grid_xy(
    farm_results,
    resolution,
    xmin=None,
    ymin=None,
    xmax=None,
    ymax=None,
    z=None,
    xspace=500.0,
    yspace=500.0,
    verbosity=0,
):
    """
    Helper function that generates 2D grid in a horizontal xy-plane.

    Parameters
    ----------
    farm_results: xarray.Dataset
        The farm results. The calculated variables have
        dimensions (state, turbine)
    resolution: float
        The resolution in m
    xmin: float
        The min x coordinate, or None for automatic
    ymin: float
        The min y coordinate, or None for automatic
    xmax: float
        The max x coordinate, or None for automatic
    ymax: float
        The max y coordinate, or None for automatic
    z: float
        The z coordinate of the plane
    xspace: float
        The extra space in x direction, before and after wind farm
    yspace: float
        The extra space in y direction, before and after wind farm
    verbosity: int, optional
        The verbosity level

    Returns
    -------
    x_pos: numpy.ndarray
        The x grid positions, shape: (n_x,)
    y_pos: numpy.ndarray
        The y grid positions, shape: (n_y,)
    z_pos: float
        The z position of the grid
    g_pts: numpy.ndarray
        The grid points, shape: (n_states, n_pts, 3)

    """

    # prepare:
    n_states = farm_results[FV.H].shape[0]

    # get base rectangle:
    x_min = xmin if xmin is not None else farm_results[FV.X].min().to_numpy() - xspace
    y_min = ymin if ymin is not None else farm_results[FV.Y].min().to_numpy() - yspace
    z_min = z if z is not None else farm_results[FV.H].min().to_numpy()
    x_max = xmax if xmax is not None else farm_results[FV.X].max().to_numpy() + xspace
    y_max = ymax if ymax is not None else farm_results[FV.Y].max().to_numpy() + yspace
    z_max = z if z is not None else farm_results[FV.H].max().to_numpy()

    x_pos, x_res = np.linspace(
        x_min,
        x_max,
        num=int((x_max - x_min) / resolution) + 1,
        endpoint=True,
        retstep=True,
        dtype=None,
    )
    y_pos, y_res = np.linspace(
        y_min,
        y_max,
        num=int((y_max - y_min) / resolution) + 1,
        endpoint=True,
        retstep=True,
        dtype=None,
    )
    N_x, N_y = len(x_pos), len(y_pos)
    n_pts = len(x_pos) * len(y_pos)
    z_pos = 0.5 * (z_min + z_max)
    g_pts = np.zeros((n_states, N_x, N_y, 3), dtype=FC.DTYPE)
    g_pts[:, :, :, 0] = x_pos[None, :, None]
    g_pts[:, :, :, 1] = y_pos[None, None, :]
    g_pts[:, :, :, 2] = z_pos

    if verbosity > 1:
        print("\nFlowPlots2D plot grid:")
        print("Min XYZ  =", x_min, y_min, z_min)
        print("Max XYZ  =", x_max, y_max, z_max)
        print("Pos Z    =", z_pos)
        print("Res XY   =", x_res, y_res)
        print("Dim XY   =", N_x, N_y)
        print("Grid pts =", n_pts)

    return (
        x_pos,
        y_pos,
        z_pos,
        g_pts.reshape(n_states, n_pts, 3),
    )


def get_grid_xz(
    farm_results,
    resolution,
    x_direction=270,
    xmin=None,
    zmin=0.0,
    xmax=None,
    zmax=None,
    y=None,
    xspace=500.0,
    zspace=500.0,
    verbosity=0,
):
    """
    Helper function that generates 2D grid in a vertical xz-plane.

    Parameters
    ----------
    farm_results: xarray.Dataset
        The farm results. The calculated variables have
        dimensions (state, turbine)
    resolution: float
        The resolution in m
    x_direction: float
        The direction of the x axis, 0 = north
    xmin: float
        The min x coordinate, or None for automatic
    zmin: float
        The min z coordinate
    xmax: float
        The max x coordinate, or None for automatic
    zmax: float
        The max z coordinate, or None for automatic
    y: float
        The y coordinate of the plane
    xspace: float
        The extra space in x direction, before and after wind farm
    zspace: float
        The extra space in z direction, below and above wind farm
    verbosity: int, optional
        The verbosity level

    Returns
    -------
    x_pos: numpy.ndarray
        The x grid positions, shape: (n_x,)
    y_pos: float
        The y position of the grid
    z_pos: numpy.ndarray
        The z grid positions, shape: (n_z,)
    g_pts: numpy.ndarray
        The grid points, shape: (n_states, n_pts, 3)

    """

    # prepare:
    n_states, n_turbines = farm_results[FV.H].shape
    n_x = np.append(wd2uv(x_direction), [0.0], axis=0)
    n_z = np.array([0.0, 0.0, 1.0])
    n_y = np.cross(n_z, n_x)

    # project to axes:
    xyz = np.zeros((n_states, n_turbines, 3), dtype=FC.DTYPE)
    xyz[:, :, 0] = farm_results[FV.X]
    xyz[:, :, 1] = farm_results[FV.Y]
    xyz[:, :, 2] = farm_results[FV.H]
    xx = np.einsum("std,d->st", xyz, n_x)
    yy = np.einsum("std,d->st", xyz, n_y)
    zz = np.einsum("std,d->st", xyz, n_z)
    del xyz

    # get base rectangle:
    x_min = xmin if xmin is not None else np.min(xx) - xspace
    z_min = zmin if zmin is not None else np.minimum(np.min(zz) - zspace, 0.0)
    y_min = y if y is not None else np.min(yy)
    x_max = xmax if xmax is not None else np.max(xx) + xspace
    z_max = zmax if zmax is not None else np.max(zz) + zspace
    y_max = y if y is not None else np.max(yy)
    del xx, yy, zz

    x_pos, x_res = np.linspace(
        x_min,
        x_max,
        num=int((x_max - x_min) / resolution) + 1,
        endpoint=True,
        retstep=True,
        dtype=None,
    )
    z_pos, z_res = np.linspace(
        z_min,
        z_max,
        num=int((z_max - z_min) / resolution) + 1,
        endpoint=True,
        retstep=True,
        dtype=None,
    )
    N_x, N_z = len(x_pos), len(z_pos)
    n_pts = len(x_pos) * len(z_pos)
    y_pos = 0.5 * (y_min + y_max)
    g_pts = np.zeros((n_states, N_x, N_z, 3), dtype=FC.DTYPE)
    g_pts[:] += x_pos[None, :, None, None] * n_x[None, None, None, :]
    g_pts[:] += y_pos * n_y[None, None, None, :]
    g_pts[:] += z_pos[None, None, :, None] * n_z[None, None, None, :]

    if verbosity > 1:
        print("\nFlowPlots2D plot grid:")
        print("Min XYZ  =", x_min, y_min, z_min)
        print("Max XYZ  =", x_max, y_max, z_max)
        print("Pos Y    =", y_pos)
        print("Res XZ   =", x_res, z_res)
        print("Dim XZ   =", N_x, N_z)
        print("Grid pts =", n_pts)

    return (
        x_pos,
        y_pos,
        z_pos,
        g_pts.reshape(n_states, n_pts, 3),
    )


def get_grid_yz(
    farm_results,
    resolution,
    x_direction=270,
    ymin=None,
    zmin=0.0,
    ymax=None,
    zmax=None,
    x=None,
    yspace=500.0,
    zspace=500.0,
    verbosity=0,
):
    """
    Helper function that generates 2D grid in a vertical yz-plane.

    Parameters
    ----------
    farm_results: xarray.Dataset
        The farm results. The calculated variables have
        dimensions (state, turbine)
    resolution: float
        The resolution in m
    x_direction: float
        The direction of the x axis, 0 = north
    ymin: float
        The min y coordinate, or None for automatic
    zmin: float
        The min z coordinate
    ymax: float
        The max y coordinate, or None for automatic
    zmax: float
        The max z coordinate, or None for automatic
    x: float
        The x coordinate of the plane
    yspace: float
        The extra space in y direction, before and after wind farm
    zspace: float
        The extra space in z direction, below and above wind farm
    verbosity: int, optional
        The verbosity level

    Returns
    -------
    x_pos: float
        The x position of the grid
    y_pos: numpy.ndarray
        The y grid positions, shape: (n_y,)
    z_pos: numpy.ndarray
        The z grid positions, shape: (n_z,)
    g_pts: numpy.ndarray
        The grid points, shape: (n_states, n_pts, 3)

    """

    # prepare:
    n_states, n_turbines = farm_results[FV.H].shape
    n_x = np.append(wd2uv(x_direction), [0.0], axis=0)
    n_z = np.array([0.0, 0.0, 1.0])
    n_y = np.cross(n_z, n_x)

    # project to axes:
    xyz = np.zeros((n_states, n_turbines, 3), dtype=FC.DTYPE)
    xyz[:, :, 0] = farm_results[FV.X]
    xyz[:, :, 1] = farm_results[FV.Y]
    xyz[:, :, 2] = farm_results[FV.H]
    xx = np.einsum("std,d->st", xyz, n_x)
    yy = np.einsum("std,d->st", xyz, n_y)
    zz = np.einsum("std,d->st", xyz, n_z)
    del xyz

    # get base rectangle:
    y_min = ymin if ymin is not None else np.min(yy) - yspace
    z_min = zmin if zmin is not None else np.minimum(np.min(zz) - zspace, 0.0)
    x_min = x if x is not None else np.min(xx)
    y_max = ymax if ymax is not None else np.max(yy) + yspace
    z_max = zmax if zmax is not None else np.max(zz) + zspace
    x_max = x if x is not None else np.max(xx)
    del xx, yy, zz

    y_pos, y_res = np.linspace(
        y_min,
        y_max,
        num=int((y_max - y_min) / resolution) + 1,
        endpoint=True,
        retstep=True,
        dtype=None,
    )
    z_pos, z_res = np.linspace(
        z_min,
        z_max,
        num=int((z_max - z_min) / resolution) + 1,
        endpoint=True,
        retstep=True,
        dtype=None,
    )
    N_y, N_z = len(y_pos), len(z_pos)
    n_pts = len(y_pos) * len(z_pos)
    x_pos = 0.5 * (x_min + x_max)
    g_pts = np.zeros((n_states, N_y, N_z, 3), dtype=FC.DTYPE)
    g_pts[:] += x_pos * n_x[None, None, None, :]
    g_pts[:] += y_pos[None, :, None, None] * n_y[None, None, None, :]
    g_pts[:] += z_pos[None, None, :, None] * n_z[None, None, None, :]

    if verbosity > 1:
        print("\nFlowPlots2D plot grid:")
        print("Min XYZ  =", x_min, y_min, z_min)
        print("Max XYZ  =", x_max, y_max, z_max)
        print("Pos X    =", x_pos)
        print("Res YZ   =", y_res, z_res)
        print("Dim YZ   =", N_y, N_z)
        print("Grid pts =", n_pts)

    return (
        x_pos,
        y_pos,
        z_pos,
        g_pts.reshape(n_states, n_pts, 3),
    )


def np2np_p(data, a_pos, b_pos):
    """
    Create numpy data from numpy data

    Parameters
    ----------
    data: dict
        The data on the grid. Key: variable name,
        value: numpy.ndarray with shape (n_gpts,)
    a_pos: numpy.ndarray
        The first axis coordinates, e.g. x_pos,
        shape: (n_a,)
    b_pos: numpy.ndarray
        The second axis coordinates, e.g. y_pos,
        shape: (n_b,)

    Returns
    -------
    out: dict
        The data on the grid. Key: variable name,
        value: numpy.ndarray with shape (n_a, n_b, n_vars)

    """
    n_a = len(a_pos)
    n_b = len(b_pos)
    n_v = len(data)
    out = np.zeros((n_a, n_b, n_v), dtype=FC.DTYPE)
    for vi, (v, d) in enumerate(data.items()):
        out[:, :, vi] = d.reshape(n_a, n_b)
    return out


def np2np_sp(data, states, a_pos, b_pos):
    """
    Create numpy data from numpy data

    Parameters
    ----------
    data: dict
        The data on the grid. Key: variable name,
        value: numpy.ndarray with shape (n_states, n_gpts)
    states: numpy.ndarray
        The states index, shape: (n_states,)
    a_pos: numpy.ndarray
        The first axis coordinates, e.g. x_pos,
        shape: (n_a,)
    b_pos: numpy.ndarray
        The second axis coordinates, e.g. y_pos,
        shape: (n_b,)

    Returns
    -------
    out: dict
        The data on the grid. Key: variable name,
        value: numpy.ndarray with shape (n_states, n_a, n_b, n_vars)

    """
    n_s = len(states)
    n_a = len(a_pos)
    n_b = len(b_pos)
    n_v = len(data)
    out = np.zeros((n_s, n_a, n_b, n_v), dtype=FC.DTYPE)
    for vi, (v, d) in enumerate(data.items()):
        out[:, :, :, vi] = d.reshape(n_s, n_a, n_b)
    return out


def np2pd_p(data, a_pos, b_pos, ori, label_map={}):
    """
    Create pandas DataFrame from numpy data

    Parameters
    ----------
    data: dict
        The data on the grid. Key: variable name,
        value: numpy.ndarray with shape (n_gpts,)
    a_pos: numpy.ndarray
        The first axis coordinates, e.g. x_pos,
        shape: (n_a,)
    b_pos: numpy.ndarray
        The second axis coordinates, e.g. y_pos,
        shape: (n_b,)
    ori: str
        The orientation, 'xy' or 'xz' or 'yz'
    label_map: dict
        The mapping from original to new field names

    Returns
    -------
    out: pandas.DataFrame
        The multi-indexed DataFrame object

    """
    a, b = [label_map.get(o, o) for o in ori]
    n_a = len(a_pos)
    n_b = len(b_pos)
    minds = pd.MultiIndex.from_product([range(n_a), range(n_b)], names=[a, b])
    return pd.DataFrame(index=minds, data=data)


def np2pd_sp(data, states, a_pos, b_pos, ori, label_map={}):
    """
    Create pandas DataFrame from numpy data

    Parameters
    ----------
    data: dict
        The data on the grid. Key: variable name,
        value: numpy.ndarray with shape (n_states, n_gpts,)
    states: numpy.ndarray
        The states index, shape: (n_states,)
    a_pos: numpy.ndarray
        The first axis coordinates, e.g. x_pos,
        shape: (n_a,)
    b_pos: numpy.ndarray
        The second axis coordinates, e.g. y_pos,
        shape: (n_b,)
    ori: str
        The orientation, 'xy' or 'xz' or 'yz'
    label_map: dict
        The mapping from original to new field names

    Returns
    -------
    out: pandas.DataFrame
        The multi-indexed DataFrame object

    """
    a, b = [label_map.get(o, o) for o in ori]
    s = label_map.get(FC.STATE, FC.STATE)
    n_a = len(a_pos)
    n_b = len(b_pos)
    minds = pd.MultiIndex.from_product(
        [states, range(n_a), range(n_b)], names=[s, a, b]
    )
    return pd.DataFrame(index=minds, data=data)


def np2xr_p(data, a_pos, b_pos, c_pos, ori, label_map={}):
    """
    Create xarray Dataset from numpy data

    Parameters
    ----------
    data: dict
        The data on the grid. Key: variable name,
        value: numpy.ndarray with shape (n_gpts,)
    a_pos: numpy.ndarray
        The first axis coordinates, e.g. x_pos,
        shape: (n_a,)
    b_pos: numpy.ndarray
        The second axis coordinates, e.g. y_pos,
        shape: (n_b,)
    ori: str
        The orientation, 'xy' or 'xz' or 'yz'
    label_map: dict
        The mapping from original to new field names

    Returns
    -------
    out: xarray.Dataset
        The Dataset object

    """
    a, b = [label_map.get(o, o) for o in ori]
    c = list(set("xyz") - set(ori))[0]
    c = label_map.get(c, c)
    n_a = len(a_pos)
    n_b = len(b_pos)
    return Dataset(
        coords={b: b_pos, a: a_pos},
        data_vars={
            v: ((b, a), np.swapaxes(d.reshape(n_a, n_b), 0, 1)) for v, d in data.items()
        },
        attrs={c: float(c_pos)},
    )


def np2xr_sp(data, states, a_pos, b_pos, c_pos, ori, label_map={}):
    """
    Create xarray Dataset from numpy data

    Parameters
    ----------
    data: dict
        The data on the grid. Key: variable name,
        value: numpy.ndarray with shape (n_states, n_gpts,)
    states: numpy.ndarray
        The states index, shape: (n_states,)
    a_pos: numpy.ndarray
        The first axis coordinates, e.g. x_pos,
        shape: (n_a,)
    b_pos: numpy.ndarray
        The second axis coordinates, e.g. y_pos,
        shape: (n_b,)
    ori: str
        The orientation, 'xy' or 'xz' or 'yz'
    label_map: dict
        The mapping from original to new field names

    Returns
    -------
    out: xarray.Dataset
        The Dataset object

    """
    a, b = [label_map.get(o, o) for o in ori]
    c = list(set("xyz") - set(ori))[0]
    c = label_map.get(c, c)
    s = label_map.get(FC.STATE, FC.STATE)
    n_s = len(states)
    n_a = len(a_pos)
    n_b = len(b_pos)
    return Dataset(
        coords={s: states, b: b_pos, a: a_pos},
        data_vars={
            v: ((s, b, a), np.swapaxes(d.reshape(n_s, n_a, n_b), 1, 2))
            for v, d in data.items()
        },
        attrs={c: float(c_pos)},
    )


def data2xr(
    x_pos,
    y_pos,
    z_pos,
    point_results,
    vars=None,
    state_mean=False,
    to_file=None,
    **kwargs,
):
    """
    Converts the image data to xarray data

    Parameter
    ---------
    x_pos: numpy.ndarray or float
        The x grid positions, shape: (n_x, 3)
    y_pos: numpy.ndarray or float
        The y grid positions, shape: (n_y, 3)
    z_pos: numpy.ndarray or float
        The z grid positions, shape: (n_z, 3)
    point_results: xarray.Dataset
        Results of calc_points
    vars: list of str, optional
        Variable selection, or None for all
    state_mean: numpy.ndarray or bool
        Computes mean over states, optionally with
        given weights
    round: dict, optional
        Round variables to given digits, or 'auto'
        for default
    to_file: str, optional
        Write to nc file
    kwargs: dict, optional
        Additional parameters for write_nc

    Returns
    -------
    ds: xarray.Dataset
        The xarray data object

    """
    if vars is None:
        vars = list(point_results.data_vars.keys())
    data = {}
    for v in vars:
        if isinstance(state_mean, np.ndarray):
            data[v] = np.einsum("sp,s->p", point_results[v].to_numpy(), state_mean)
        elif state_mean:
            data[v] = np.mean(point_results[v].to_numpy(), axis=0)
        else:
            data[v] = point_results[v].to_numpy()

    allc = [x_pos, y_pos, z_pos]
    allcn = ["x", "y", "z"]
    ci = [i for i, x in enumerate(allc) if isinstance(x, np.ndarray)]
    cj = [i for i in range(3) if i not in ci][0]
    cl = [len(allc[i]) for i in ci]
    cn = list(reversed([allcn[i] for i in ci]))

    coords = {}
    attrs = {allcn[cj]: allc[cj].to_numpy()}
    if (
        FC.STATE in point_results.coords
        and isinstance(state_mean, bool)
        and not state_mean
    ):
        if point_results.sizes[FC.STATE] > 1:
            coords[FC.STATE] = point_results[FC.STATE].to_numpy()
        else:
            attrs[FC.STATE] = str(point_results[FC.STATE][0].to_numpy())
    coords.update({allcn[i]: allc[i] for i in reversed(ci)})

    dvars = {}
    for v, d in data.items():
        if len(d.shape) == 1:
            dvars[v] = (cn, np.swapaxes(d.reshape(*cl), 0, 1))
        else:
            dvars[v] = ([FC.STATE] + cn, np.swapaxes(d.reshape(d.shape[0], *cl), 1, 2))

    ds = Dataset(coords=coords, data_vars=dvars, attrs=attrs)

    if to_file is not None:
        write_nc(ds, to_file, **kwargs)

    return ds



import numpy as np

from foxes.utils import Dict
import foxes.variables as FV
import foxes.constants as FC


class Data(Dict):
    """
    Container for numpy array data and
    the associated meta data.

    Attributes
    ----------
    dims: dict
        The dimensions tuples, same or subset
        of data keys
    loop_dims: array_like of str
        List of the loop dimensions during xarray's
        `apply_ufunc` calculations
    sizes: dict
        The dimension sizes

    :group: core

    """

    def __init__(self, data, dims, loop_dims, name="data"):
        """
        Constructor.

        Parameters
        ----------
        data: dict
            The initial data to be stored
        dims: dict
            The dimensions tuples, same or subset
            of data keys
        loop_dims: array_like of str
            List of the loop dimensions during xarray's
            `apply_ufunc` calculations
        name: str
            The data container name

        """
        super().__init__(name=name)

        self.update(data)
        self.dims = dims
        self.loop_dims = loop_dims

        self.sizes = {}
        for v, d in data.items():
            self._run_entry_checks(v, d, dims[v])

        self._auto_update()

    @property
    def n_states(self):
        """
        The number of states

        Returns
        -------
        int:
            The number of states

        """
        return self.sizes[FC.STATE] if FC.STATE in self.sizes else None

    @property
    def n_turbines(self):
        """
        The number of turbines

        Returns
        -------
        int:
            The number of turbines

        """
        return self.sizes[FC.TURBINE] if FC.TURBINE in self.sizes else None

    def states_i0(self, counter=False, algo=None):
        """
        Get the state counter for first state in chunk

        Parameters
        ----------
        counter: bool
            Return the state counter instead of the index
        algo: foxes.core.Algorithm, optional
            The algorithm, required for state counter

        Returns
        -------
        int:
            The state counter for first state in chunk
            or the corresponding index

        """
        if FC.STATE not in self:
            return None
        elif counter:
            if algo is None:
                raise KeyError(f"{self.name}: Missing algo for deducing state counter")
            return np.argwhere(algo.states.index() == self[FC.STATE][0])[0][0]
        else:
            return self[FC.STATE][0]

    def _auto_update(self):
        """Checks and operations after data changes"""
        data = self
        dims = self.dims

        if (
            FV.TXYH not in data
            and FV.X in data
            and FV.Y in data
            and FV.H in data
            and dims[FV.X] == dims[FV.Y]
            and dims[FV.X] == dims[FV.H]
        ):
            self[FV.TXYH] = np.stack([self[FV.X], self[FV.Y], self[FV.H]], axis=-1)

            self[FV.X] = self[FV.TXYH][..., 0]
            self[FV.Y] = self[FV.TXYH][..., 1]
            self[FV.H] = self[FV.TXYH][..., 2]

            self.dims[FV.TXYH] = tuple(list(dims[FV.X]) + [FC.XYH])

    def _run_entry_checks(self, name, data, dims):
        """Run entry checks on new data"""
        # remove axes of size 1, added by dask for extra loop dimensions:
        if dims is not None:
            if len(dims) != len(data.shape):
                for li, l in enumerate(self.loop_dims):
                    if data.shape[li] == 1 and (len(dims) < li + 1 or dims[li] != l):
                        self[name] = np.squeeze(data, axis=li)

            for ci, c in enumerate(dims):
                if c not in self.sizes:
                    self.sizes[c] = self[name].shape[ci]
                elif self.sizes[c] != self[name].shape[ci]:
                    raise ValueError(
                        f"Inconsistent size for data entry '{name}', dimension '{c}': Expecting {self.sizes[c]}, found {self[name].shape[ci]} in shape {self[name].shape}"
                    )

    def add(self, name, data, dims):
        """
        Add data entry

        Parameters
        ----------
        name: str
            The data name
        data: np.ndarray
            The data
        dims: tuple of str
            The dimensions

        """
        self[name] = data
        self.dims[name] = dims
        self._run_entry_checks(name, data, dims)
        self._auto_update()

    def get_slice(self, s, dim_map={}, name=None, keep=True):
        """
        Get a slice of data.

        Parameters
        ----------
        s: slice
            The slice
        dim_map: dict
            Mapping from original to new dimensions.
            If not found, same dimensions are assumed.
        name: str, optional
            The name of the data object
        keep: bool
            Keep non-matching fields as they are, else
            throw them out

        Returns
        -------
        data: Data
            The new data object, containing slices

        """
        data = {}
        dims = {}
        for v in self.keys():
            try:
                d = self.dims[v]
                data[v] = self[v][s]
                dims[v] = dim_map.get(d, d)
            except IndexError:
                if keep:
                    data[v] = self[v]
                    dims[v] = self.dims[v]
        if name is None:
            name = self.name
        return type(self)(data, dims, loop_dims=self.loop_dims, name=name)


class MData(Data):
    """
    Container for foxes model data.

    :group: core

    """

    def __init__(self, *args, name="mdata", **kwargs):
        """
        Constructor

        Parameters
        ----------
        args: tuple, optional
            Arguments for the base class
        name: str
            The data name
        kwargs: dict, optional
            Arguments for the base class

        """
        super().__init__(*args, name=name, **kwargs)


class FData(Data):
    """
    Container for foxes farm data.

    Each farm data entry has (n_states, n_turbines) shape,
    except the dimensions.

    :group: core

    """

    def __init__(self, *args, name="fdata", **kwargs):
        """
        Constructor

        Parameters
        ----------
        args: tuple, optional
            Arguments for the base class
        name: str
            The data name
        kwargs: dict, optional
            Arguments for the base class

        """
        super().__init__(*args, name=name, **kwargs)

    def _run_entry_checks(self, name, data, dims):
        """Run entry checks on new data"""
        super()._run_entry_checks(name, data, dims)
        data = self[name]
        dims = self.dims[name]

        if name not in self.sizes and name not in FC.TNAME:
            dms = (FC.STATE, FC.TURBINE)
            shp = (self.n_states, self.n_turbines)
            if len(data.shape) < 2:
                raise ValueError(
                    f"FData '{self.name}': Invalid shape for '{name}', expecting {shp}, got {data.shape}"
                )
            if len(dims) < 2 or dims[:2] != dms:
                raise ValueError(
                    f"FData '{self.name}': Invalid dims for '{name}', expecting {dms}, got {dims}"
                )

    def _auto_update(self):
        """Checks and operations after data changes"""
        super()._auto_update()
        if len(self):
            for x in [FC.STATE, FC.TURBINE]:
                if x not in self.sizes:
                    raise KeyError(
                        f"FData '{self.name}': Missing '{x}' in sizes, got {sorted(list(self.sizes.keys()))}"
                    )


class TData(Data):
    """
    Container for foxes target data.

    Each target consists of a fixed number of
    target points.

    :group: core

    """

    def __init__(self, *args, name="tdata", **kwargs):
        """
        Constructor

        Parameters
        ----------
        args: tuple, optional
            Arguments for the base class
        name: str
            The data name
        kwargs: dict, optional
            Arguments for the base class

        """
        super().__init__(*args, name=name, **kwargs)

    def _run_entry_checks(self, name, data, dims):
        """Run entry checks on new data"""
        super()._run_entry_checks(name, data, dims)
        data = self[name]
        dims = self.dims[name]

        if name == FC.TARGETS:
            dms = (FC.STATE, FC.TARGET, FC.TPOINT, FC.XYH)
            shp = (self.n_states, self.n_targets, self.n_tpoints, 3)
            if dims != dms:
                raise ValueError(
                    f"TData '{self.name}': Invalid dims of {FC.TARGETS}, expecting {dms}, got {dims}"
                )
            if data.shape != shp:
                raise ValueError(
                    f"TData '{self.name}': Invalid shape of {FC.TARGETS}, expecting {shp}, got {data.shape}"
                )

        elif name == FC.TWEIGHTS:
            dms = (FC.TPOINT,)
            shp = (self.n_tpoints,)
            if dims != dms:
                raise ValueError(
                    f"TData '{self.name}': Invalid dims of {FC.TWEIGHTS}, expecting {dms}, got {dims}"
                )
            if data.shape != shp:
                raise ValueError(
                    f"TData '{self.name}': Invalid shape of {FC.TWEIGHTS}, expecting {shp}, got {data.shape}"
                )

        elif FC.TARGETS not in self:
            raise KeyError(
                f"TData '{self.name}': Missing '{FC.TARGETS}' before adding '{name}'"
            )

        elif FC.TWEIGHTS not in self:
            raise KeyError(
                f"TData '{self.name}': Missing '{FC.TWEIGHTS}' before adding '{name}'"
            )

        elif name not in self.sizes:
            dms = (FC.STATE, FC.TARGET, FC.TPOINT)
            shp = (self.n_states, self.n_targets, self.n_tpoints)
            if len(data.shape) < 3:
                raise ValueError(
                    f"TData '{self.name}': Invalid shape for '{name}', expecting {shp}, got {data.shape}"
                )
            if len(dims) < 3 or dims[:3] != dms:
                raise ValueError(
                    f"TData '{self.name}': Invalid dims for '{name}', expecting {dms}, got {dims}"
                )

    def _auto_update(self):
        """Checks and operations after data changes"""
        super()._auto_update()
        if len(self):
            for x in [FC.TARGETS, FC.TWEIGHTS]:
                if x not in self:
                    raise KeyError(
                        f"TData '{self.name}': Missing '{x}' in data, got {sorted(list(self.keys()))}"
                    )
                if x not in self.dims:
                    raise KeyError(
                        f"TData '{self.name}': Missing '{x}' in dims, got {sorted(list(self.dims.keys()))}"
                    )
            for x in [FC.STATE, FC.TARGET, FC.TPOINT]:
                if x not in self.sizes:
                    raise KeyError(
                        f"TData '{self.name}': Missing '{x}' in sizes, got {sorted(list(self.sizes.keys()))}"
                    )

    @property
    def n_targets(self):
        """
        The number of targets

        Returns
        -------
        int:
            The number of targets

        """
        return self.sizes[FC.TARGET]

    @property
    def n_tpoints(self):
        """
        The number of points per target

        Returns
        -------
        int:
            The number of points per target

        """
        return self.sizes[FC.TPOINT]

    def tpoint_mean(self, variable):
        """
        Take the mean over target points

        Parameters
        ----------
        variable: str
            The variable name

        Returns
        -------
        data: numpy.ndarray
            The reduced array, shape:
            (n_states, n_targets, ...)

        """
        return np.einsum("stp...,p->st...", self[variable], self[FC.TWEIGHTS])

    @classmethod
    def from_points(
        cls,
        points,
        data={},
        dims={},
        name="tdata",
        **kwargs,
    ):
        """
        Create from points

        Parameters
        ----------
        points: np.ndarray
            The points, shape: (n_states, n_points, 3)
        data: dict
            The initial data to be stored
        dims: dict
            The dimensions tuples, same or subset
            of data keys
        name: str
            The data container name
        kwargs: dict, optional
            Additional parameters for the constructor

        Returns
        -------
        pdata: Data
            The data object

        """
        if len(points.shape) != 3 or points.shape[2] != 3:
            raise ValueError(
                f"Expecting points shape (n_states, n_points, 3), got {points.shape}"
            )
        data[FC.TARGETS] = points[:, :, None, :]
        dims[FC.TARGETS] = (FC.STATE, FC.TARGET, FC.TPOINT, FC.XYH)
        data[FC.TWEIGHTS] = np.array([1], dtype=FC.DTYPE)
        dims[FC.TWEIGHTS] = (FC.TPOINT,)
        return cls(data, dims, [FC.STATE, FC.TARGET], name=name, **kwargs)

    @classmethod
    def from_tpoints(
        cls,
        tpoints,
        tweights,
        data={},
        dims={},
        name="tdata",
        **kwargs,
    ):
        """
        Create from points at targets

        Parameters
        ----------
        tpoints: np.ndarray
            The points at targets, shape:
            (n_states, n_targets, n_tpoints, 3)
        tweights: np.ndarray, optional
            The target point weights, shape:
            (n_tpoints,)
        data: dict
            The initial data to be stored
        dims: dict
            The dimensions tuples, same or subset
            of data keys
        name: str
            The data container name
        kwargs: dict, optional
            Additional parameters for the constructor

        Returns
        -------
        pdata: Data
            The data object

        """
        if len(tpoints.shape) != 4 or tpoints.shape[3] != 3:
            raise ValueError(
                f"Expecting tpoints shape (n_states, n_targets, n_tpoints, 3), got {tpoints.shape}"
            )
        data[FC.TARGETS] = tpoints
        dims[FC.TARGETS] = (FC.STATE, FC.TARGET, FC.TPOINT, FC.XYH)
        data[FC.TWEIGHTS] = tweights
        dims[FC.TWEIGHTS] = (FC.TPOINT,)
        return cls(data, dims, [FC.STATE], name=name, **kwargs)



import numpy as np
from abc import abstractmethod

from foxes.models.wake_models.dist_sliced import DistSlicedWakeModel


class AxisymmetricWakeModel(DistSlicedWakeModel):
    """
    Abstract base class for wake models
    that depend on (x, r) separately.

    The ability to evaluate multiple r values per x
    is used by the `PartialAxiwake` partial wakes model.

    :group: models.wake_models

    """

    @abstractmethod
    def calc_wakes_x_r(
        self,
        algo,
        mdata,
        fdata,
        tdata,
        downwind_index,
        x,
        r,
    ):
        """
        Calculate wake deltas.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        mdata: foxes.core.MData
            The model data
        fdata: foxes.core.FData
            The farm data
        tdata: foxes.core.TData
            The target point data
        downwind_index: int
            The index in the downwind order
        x: numpy.ndarray
            The x values, shape: (n_states, n_targets)
        r: numpy.ndarray
            The radial values for each x value, shape:
            (n_states, n_targets, n_yz_per_target)

        Returns
        -------
        wdeltas: dict
            The wake deltas. Key: variable name str,
            value: numpy.ndarray, shape: (n_st_sel, n_r_per_x)
        st_sel: numpy.ndarray of bool
            The state-target selection, for which the wake
            is non-zero, shape: (n_states, n_targets)

        """
        pass

    def calc_wakes_x_yz(
        self,
        algo,
        mdata,
        fdata,
        tdata,
        downwind_index,
        x,
        yz,
    ):
        """
        Calculate wake deltas.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        mdata: foxes.core.MData
            The model data
        fdata: foxes.core.FData
            The farm data
        tdata: foxes.core.TData
            The target point data
        downwind_index: int
            The index in the downwind order
        x: numpy.ndarray
            The x values, shape: (n_states, n_targets)
        yz: numpy.ndarray
            The yz values for each x value, shape:
            (n_states, n_targets, n_yz_per_target, 2)

        Returns
        -------
        wdeltas: dict
            The wake deltas. Key: variable name str,
            value: numpy.ndarray, shape: (n_st_sel, n_yz_per_target)
        st_sel: numpy.ndarray of bool
            The state-target selection, for which the wake
            is non-zero, shape: (n_states, n_targets)

        """
        r = np.linalg.norm(yz, axis=-1)
        return self.calc_wakes_x_r(algo, mdata, fdata, tdata, downwind_index, x, r)



import numpy as np

from foxes.core import RotorModel
import foxes.constants as FC


class GridRotor(RotorModel):
    """
    The weighted regular grid rotor model, composed maximally
    of n x n points, possibly kicking out the outside points.

    Attributes
    ----------
    n: int
        The number of points along one direction,
        maximal number of points is N = n * n
    reduce: bool
        Flag for reduction to points actually representing
        an area with overlap with the circe, recalculating
        the self.weights accordingly
    nint: int
        Integration steps per element

    :group: models.rotor_models

    """

    def __init__(self, n, calc_vars, reduce=True, nint=200):
        """
        Constructor.

        Parameters
        ----------
        n: int
            The number of points along one direction,
            maximal number of points is N = n * n
        calc_vars: list of str
            The variables that are calculated by the model
            (Their ambients are added automatically)
        reduce: bool
            Flag for reduction to points actually representing
            an area with overlap with the circe, recalculating
            the self.weights accordingly
        nint: int
            Integration steps per element
        name: str, optional
            The model name

        """
        super().__init__(calc_vars)

        self.n = n
        self.reduce = reduce
        self.nint = nint

    def __repr__(self):
        r = "" if self.reduce else ", reduce=False"
        return f"{type(self).__name__}(n={self.n}){r}"

    def initialize(self, algo, verbosity=0):
        """
        Initializes the model.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        verbosity: int
            The verbosity level, 0 = silent

        """
        super().initialize(algo, verbosity)

        N = self.n * self.n
        delta = 2.0 / self.n
        x = [-1.0 + (i + 0.5) * delta for i in range(self.n)]
        x, y = np.meshgrid(x, x, indexing="ij")

        self.dpoints = np.zeros([N, 3], dtype=FC.DTYPE)
        self.dpoints[:, 1] = x.reshape(N)
        self.dpoints[:, 2] = y.reshape(N)

        if self.reduce:
            self.weights = np.zeros((self.n, self.n), dtype=FC.DTYPE)
            for i in range(0, self.n):
                for j in range(0, self.n):
                    d = delta / self.nint
                    hx = [
                        x[i, j] - delta / 2.0 + (k + 0.5) * d for k in range(self.nint)
                    ]
                    hy = [
                        y[i, j] - delta / 2.0 + (k + 0.5) * d for k in range(self.nint)
                    ]
                    pts = np.zeros((self.nint, self.nint, 2), dtype=FC.DTYPE)
                    pts[:, :, 0], pts[:, :, 1] = np.meshgrid(hx, hy, indexing="ij")

                    d = np.linalg.norm(pts, axis=2)
                    self.weights[i, j] = np.sum(d <= 1.0) / self.nint**2

            self.weights = self.weights.reshape(N)
            sel = self.weights > 0.0
            self.dpoints = self.dpoints[sel]
            self.weights = self.weights[sel]
            self.weights /= np.sum(self.weights)

        else:
            self.dpoints[:, 1] = x.reshape(N)
            self.dpoints[:, 2] = y.reshape(N)
            self.weights = np.ones(N, dtype=FC.DTYPE) / N

    def n_rotor_points(self):
        """
        The number of rotor points

        Returns
        -------
        n_rpoints: int
            The number of rotor points

        """
        return len(self.weights)

    def design_points(self):
        """
        The rotor model design points.

        Design points are formulated in rotor plane
        (x,y,z)-coordinates in rotor frame, such that
        - (0,0,0) is the centre point,
        - (1,0,0) is the point radius * n_rotor_axis
        - (0,1,0) is the point radius * n_rotor_side
        - (0,0,1) is the point radius * n_rotor_up

        Returns
        -------
        dpoints: numpy.ndarray
            The design points, shape: (n_points, 3)

        """
        return self.dpoints

    def rotor_point_weights(self):
        """
        The weights of the rotor points

        Returns
        -------
        weights: numpy.ndarray
            The weights of the rotor points,
            add to one, shape: (n_rpoints,)

        """
        return self.weights



import numpy as np

from foxes.core import WakeFrame
from foxes.utils import wd2uv
from foxes.core.data import MData, FData, TData
import foxes.variables as FV
import foxes.constants as FC


class Timelines(WakeFrame):
    """
    Dynamic wakes for spatially uniform timeseries states.

    Attributes
    ----------
    max_wake_length: float
        The maximal wake length
    cl_ipars: dict
        Interpolation parameters for centre line
        point interpolation
    dt_min: float, optional
        The delta t value in minutes,
        if not from timeseries data

    :group: models.wake_frames

    """

    def __init__(self, max_wake_length=2e4, cl_ipars={}, dt_min=None):
        """
        Constructor.

        Parameters
        ----------
        max_wake_length: float
            The maximal wake length
        cl_ipars: dict
            Interpolation parameters for centre line
            point interpolation
        dt_min: float, optional
            The delta t value in minutes,
            if not from timeseries data

        """
        super().__init__()
        self.max_wake_length = max_wake_length
        self.cl_ipars = cl_ipars
        self.dt_min = dt_min

    def __repr__(self):
        return f"{type(self).__name__}(dt_min={self.dt_min})"

    def initialize(self, algo, verbosity=0):
        """
        Initializes the model.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        verbosity: int
            The verbosity level, 0 = silent

        """
        super().initialize(algo, verbosity)

        if verbosity > 0:
            print(f"{self.name}: Pre-calculating ambient wind vectors")

        # get and check times:
        times = np.asarray(algo.states.index())
        if self.dt_min is None:
            if not np.issubdtype(times.dtype, np.datetime64):
                raise TypeError(
                    f"{self.name}: Expecting state index of type np.datetime64, found {times.dtype}"
                )
            elif len(times) == 1:
                raise KeyError(
                    f"{self.name}: Expecting 'dt_min' for single step timeseries"
                )
            dt = (times[1:] - times[:-1]).astype("timedelta64[s]").astype(FC.ITYPE)
        else:
            n = max(len(times) - 1, 1)
            dt = np.full(n, self.dt_min * 60, dtype="timedelta64[s]").astype(FC.ITYPE)

        # calculate horizontal wind vector in all states:
        self._uv = np.zeros((algo.n_states, 1, 3), dtype=FC.DTYPE)

        # prepare mdata:
        mdata = algo.get_model_data(algo.states)["data_vars"]
        mdict = {v: d[1] for v, d in mdata.items()}
        mdims = {v: d[0] for v, d in mdata.items()}
        mdata = MData(mdict, mdims, loop_dims=[FC.STATE])
        del mdict, mdims

        # prepare fdata:
        fdata = FData({}, {}, loop_dims=[FC.STATE])

        # prepare tdata:
        tdata = {
            v: np.zeros((algo.n_states, 1, 1), dtype=FC.DTYPE)
            for v in algo.states.output_point_vars(algo)
        }
        pdims = {v: (FC.STATE, FC.TARGET, FC.TPOINT) for v in tdata.keys()}
        tdata = TData.from_points(
            points=np.zeros((algo.n_states, 1, 3), dtype=FC.DTYPE),
            data=tdata,
            dims=pdims,
        )

        # calculate:
        res = algo.states.calculate(algo, mdata, fdata, tdata)
        if len(dt) == 1:
            self._dxy = wd2uv(res[FV.WD], res[FV.WS])[:, 0, 0, :2] * dt[:, None]
        else:
            self._dxy = wd2uv(res[FV.WD], res[FV.WS])[:-1, 0, 0, :2] * dt[:, None]
            self._dxy = np.insert(self._dxy, 0, self._dxy[0], axis=0)

        """ DEBUG
        import matplotlib.pyplot as plt
        xy = np.array([np.sum(self._dxy[:n], axis=0) for n in range(len(self._dxy))])
        print(xy)
        plt.plot(xy[:, 0], xy[:, 1])
        plt.show()
        quit()
        """

    def calc_order(self, algo, mdata, fdata):
        """ "
        Calculates the order of turbine evaluation.

        This function is executed on a single chunk of data,
        all computations should be based on numpy arrays.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        mdata: foxes.core.MData
            The model data
        fdata: foxes.core.FData
            The farm data

        Returns
        -------
        order: numpy.ndarray
            The turbine order, shape: (n_states, n_turbines)

        """
        # prepare:
        n_states = fdata.n_states
        n_turbines = algo.n_turbines
        tdata = TData.from_points(points=fdata[FV.TXYH])

        # calculate streamline x coordinates for turbines rotor centre points:
        # n_states, n_turbines_source, n_turbines_target
        coosx = np.zeros((n_states, n_turbines, n_turbines), dtype=FC.DTYPE)
        for ti in range(n_turbines):
            coosx[:, ti, :] = self.get_wake_coos(algo, mdata, fdata, tdata, ti)[
                :, :, 0, 0
            ]

        # derive turbine order:
        # TODO: Remove loop over states
        order = np.zeros((n_states, n_turbines), dtype=FC.ITYPE)
        for si in range(n_states):
            order[si] = np.lexsort(keys=coosx[si])

        return order

    def get_wake_coos(
        self,
        algo,
        mdata,
        fdata,
        tdata,
        downwind_index,
    ):
        """
        Calculate wake coordinates of rotor points.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        mdata: foxes.core.MData
            The model data
        fdata: foxes.core.FData
            The farm data
        tdata: foxes.core.TData
            The target point data
        downwind_index: int
            The index of the wake causing turbine
            in the downwnd order

        Returns
        -------
        wake_coos: numpy.ndarray
            The wake frame coordinates of the evaluation
            points, shape: (n_states, n_targets, n_tpoints, 3)

        """
        # prepare:
        targets = tdata[FC.TARGETS]
        n_states, n_targets, n_tpoints = targets.shape[:3]
        n_points = n_targets * n_tpoints
        points = targets.reshape(n_states, n_points, 3)
        rxyz = fdata[FV.TXYH][:, downwind_index]

        D = np.zeros((n_states, n_points), dtype=FC.DTYPE)
        D[:] = fdata[FV.D][:, downwind_index, None]

        i0 = mdata.states_i0(counter=True, algo=algo)
        i1 = i0 + mdata.n_states
        dxy = self._dxy[:i1]

        trace_p = np.zeros((n_states, n_points, 2), dtype=FC.DTYPE)
        trace_p[:] = points[:, :, :2] - rxyz[:, None, :2]
        trace_l = np.zeros((n_states, n_points), dtype=FC.DTYPE)
        trace_d = np.full((n_states, n_points), np.inf, dtype=FC.DTYPE)
        trace_si = np.zeros((n_states, n_points), dtype=FC.ITYPE)
        trace_si[:] = i0 + np.arange(n_states)[:, None] + 1

        wcoos = np.full((n_states, n_points, 3), 1e20, dtype=FC.DTYPE)
        wcoosx = wcoos[:, :, 0]
        wcoosy = wcoos[:, :, 1]
        wcoos[:, :, 2] = points[:, :, 2] - rxyz[:, None, 2]
        del rxyz

        while True:
            sel = (trace_si > 0) & (trace_l < self.max_wake_length)
            if np.any(sel):
                trace_si[sel] -= 1

                delta = dxy[trace_si[sel]]
                dmag = np.linalg.norm(delta, axis=-1)

                trace_p[sel] -= delta
                trace_l[sel] += dmag

                trp = trace_p[sel]
                d0 = trace_d[sel]
                d = np.linalg.norm(trp, axis=-1)
                trace_d[sel] = d

                seln = d <= np.minimum(d0, 2 * dmag)
                if np.any(seln):
                    htrp = trp[seln]
                    raxis = delta[seln]
                    raxis = raxis / np.linalg.norm(raxis, axis=-1)[:, None]
                    saxis = np.concatenate(
                        [-raxis[:, 1, None], raxis[:, 0, None]], axis=1
                    )

                    wcx = wcoosx[sel]
                    wcx[seln] = np.einsum("sd,sd->s", htrp, raxis) + trace_l[sel][seln]
                    wcoosx[sel] = wcx
                    del wcx, raxis

                    wcy = wcoosy[sel]
                    wcy[seln] = np.einsum("sd,sd->s", htrp, saxis)
                    wcoosy[sel] = wcy
                    del wcy, saxis, htrp

            else:
                break

        # turbines that cause wake:
        tdata[FC.STATE_SOURCE_ORDERI] = downwind_index

        # states that cause wake for each target point:
        tdata.add(
            FC.STATES_SEL,
            trace_si.reshape(n_states, n_targets, n_tpoints),
            (FC.STATE, FC.TARGET, FC.TPOINT),
        )

        return wcoos.reshape(n_states, n_targets, n_tpoints, 3)

    def get_centreline_points(self, algo, mdata, fdata, downwind_index, x):
        """
        Gets the points along the centreline for given
        values of x.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        mdata: foxes.core.MData
            The model data
        fdata: foxes.core.FData
            The farm data
        downwind_index: int
            The index in the downwind order
        x: numpy.ndarray
            The wake frame x coordinates, shape: (n_states, n_points)

        Returns
        -------
        points: numpy.ndarray
            The centreline points, shape: (n_states, n_points, 3)

        """
        raise NotImplementedError



import numpy as np
import xarray as xr

from .model import Model
from foxes.data import StaticData
from foxes.utils import Dict, all_subclasses
import foxes.constants as FC


class Algorithm(Model):
    """
    Abstract base class for algorithms.

    Algorithms collect required objects for running
    calculations, and contain the calculation functions
    which are meant to be called from top level code.

    Attributes
    ----------
    mbook: foxes.models.ModelBook
        The model book
    farm: foxes.WindFarm
        The wind farm
    chunks: dict
        The chunks choice for running in parallel with dask,
        e.g. `{"state": 1000}` for chunks of 1000 states
    verbosity: int
        The verbosity level, 0 means silent
    dbook: foxes.DataBook
        The data book, or None for default

    :group: core

    """

    def __init__(self, mbook, farm, chunks, verbosity, dbook=None):
        """
        Constructor.

        Parameters
        ----------
        mbook: foxes.models.ModelBook
            The model book
        farm: foxes.WindFarm
            The wind farm
        chunks: dict
            The chunks choice for running in parallel with dask,
            e.g. `{"state": 1000}` for chunks of 1000 states
        verbosity: int
            The verbosity level, 0 means silent
        dbook: foxes.DataBook, optional
            The data book, or None for default

        """
        super().__init__()

        self.name = type(self).__name__
        self.mbook = mbook
        self.farm = farm
        self.chunks = chunks
        self.verbosity = verbosity
        self.n_states = None
        self.n_turbines = farm.n_turbines
        self.dbook = StaticData() if dbook is None else dbook

        if chunks is not None and FC.TARGET not in chunks:
            self.chunks[FC.TARGET] = chunks.get(FC.POINT, None)

        self._idata_mem = Dict()

    def print(self, *args, vlim=1, **kwargs):
        """
        Print function, based on verbosity.

        Parameters
        ----------
        args: tuple, optional
            Arguments for the print function
        kwargs: dict, optional
            Keyword arguments for the print function
        vlim: int
            The verbosity limit

        """
        if self.verbosity >= vlim:
            print(*args, **kwargs)

    def __get_sizes(self, idata, mtype):
        """
        Private helper function
        """

        sizes = {}
        for v, t in idata["data_vars"].items():
            if not isinstance(t, tuple) or len(t) != 2:
                raise ValueError(
                    f"Input {mtype} data entry '{v}': Not a tuple of size 2, got '{t}'"
                )
            if not isinstance(t[0], tuple):
                raise ValueError(
                    f"Input {mtype} data entry '{v}': First tuple entry not a dimensions tuple, got '{t[0]}'"
                )
            for c in t[0]:
                if not isinstance(c, str):
                    raise ValueError(
                        f"Input {mtype} data entry '{v}': First tuple entry not a dimensions tuple, got '{t[0]}'"
                    )
            if not isinstance(t[1], np.ndarray):
                raise ValueError(
                    f"Input {mtype} data entry '{v}': Second entry is not a numpy array, got: {type(t[1]).__name__}"
                )
            if len(t[1].shape) != len(t[0]):
                raise ValueError(
                    f"Input {mtype} data entry '{v}': Wrong data shape, expecting {len(t[0])} dimensions, got {t[1].shape}"
                )
            if FC.STATE in t[0]:
                if t[0][0] != FC.STATE:
                    raise ValueError(
                        f"Input {mtype} data entry '{v}': Dimension '{FC.STATE}' not at first position, got {t[0]}"
                    )
                if FC.TURBINE in t[0]:
                    if t[0][1] != FC.TURBINE:
                        raise ValueError(
                            f"Input {mtype} data entry '{v}': Dimension '{FC.TURBINE}' not at second position, got {t[0]}"
                        )
                if FC.TARGET in t[0]:
                    if t[0][1] != FC.TARGET:
                        raise ValueError(
                            f"Input {mtype} data entry '{v}': Dimension '{FC.TARGET}' not at second position, got {t[0]}"
                        )
                    if len(t[0]) < 3 or t[0][2] != FC.TPOINT:
                        raise KeyError(
                            f"Input {mtype} data entry '{v}': Expecting dimension '{FC.TPOINT}' as third entry. Got {t[0]}"
                        )
            elif FC.TURBINE in t[0]:
                raise ValueError(
                    f"Input {mtype} data entry '{v}': Dimension '{FC.TURBINE}' requires combination with dimension '{FC.STATE}'"
                )
            for d, s in zip(t[0], t[1].shape):
                if d not in sizes:
                    sizes[d] = s
                elif sizes[d] != s:
                    raise ValueError(
                        f"Input {mtype} data entry '{v}': Dimension '{d}' has wrong size, expecting {sizes[d]}, got {s}"
                    )
        for v, c in idata["coords"].items():
            if v not in sizes:
                raise KeyError(
                    f"Input coords entry '{v}': Not used in farm data, found {sorted(list(sizes.keys()))}"
                )
            elif len(c) != sizes[v]:
                raise ValueError(
                    f"Input coords entry '{v}': Wrong coordinate size for '{v}': Expecting {sizes[v]}, got {len(c)}"
                )

        return sizes

    def __get_xrdata(self, idata, sizes):
        """
        Private helper function
        """
        xrdata = xr.Dataset(**idata)
        if self.chunks is not None:
            if FC.TURBINE in self.chunks.keys():
                raise ValueError(
                    f"Dimension '{FC.TURBINE}' cannot be chunked, got chunks {self.chunks}"
                )
            if FC.TPOINT in self.chunks.keys():
                raise ValueError(
                    f"Dimension '{FC.TPOINT}' cannot be chunked, got chunks {self.chunks}"
                )
            xrdata = xrdata.chunk(
                chunks={c: v for c, v in self.chunks.items() if c in sizes}
            )
        return xrdata

    def chunked(self, ds):
        return (
            ds.chunk(chunks={c: v for c, v in self.chunks.items() if c in ds.coords})
            if self.chunks is not None
            else ds
        )

    def initialize(self):
        """
        Initializes the algorithm.
        """
        super().initialize(self, self.verbosity)

    @property
    def idata_mem(self):
        """
        The current idata memory

        Returns
        -------
        dict :
            Keys: model name, value: idata dict

        """
        return self._idata_mem

    def store_model_data(self, model, idata, force=False):
        """
        Store model data

        Parameters
        ----------
        model: foxes.core.Model
            The model
        idata: dict
            The dict has exactly two entries: `data_vars`,
            a dict with entries `name_str -> (dim_tuple, data_ndarray)`;
            and `coords`, a dict with entries `dim_name_str -> dim_array`
        force: bool
            Overwrite existing data

        """
        mname = f"{type(model).__name__}_{model.name}"
        if not force and mname in self._idata_mem:
            raise KeyError(f"Attempt to overwrite stored data for model '{mname}'")
        self._idata_mem[mname] = idata

    def get_model_data(self, model):
        """
        Gets model data from memory

        Parameters
        ----------
        model: foxes.core.Model
            The model

        """
        mname = f"{type(model).__name__}_{model.name}"
        try:
            return self._idata_mem[mname]
        except KeyError:
            raise KeyError(
                f"Key '{mname}' not found in idata_mem, available keys: {sorted(list(self._idata_mem.keys()))}"
            )

    def del_model_data(self, model):
        """
        Remove stored model data

        Parameters
        ----------
        model: foxes.core.Model
            The model

        """
        mname = f"{type(model).__name__}_{model.name}"
        try:
            del self._idata_mem[mname]
        except KeyError:
            raise KeyError(f"Attempt to delete data of model '{mname}', but not stored")

    def update_n_turbines(self):
        """
        Reset the number of turbines,
        according to self.farm
        """
        if self.n_turbines != self.farm.n_turbines:
            self.n_turbines = self.farm.n_turbines

            # resize stored idata, if dependent on turbine coord:
            newk = {}
            for mname, idata in self.idata_mem.items():
                if mname[:2] == "__":
                    continue
                for dname, d in idata["data_vars"].items():
                    k = f"__{mname}_{dname}_turbinv"
                    if k in self.idata_mem:
                        ok = self.idata_mem[k]
                    else:
                        ok = None
                        if FC.TURBINE in d[0]:
                            i = d[0].index(FC.TURBINE)
                            ok = np.unique(d[1], axis=1).shape[i] == 1
                        newk[k] = ok
                    if ok is not None:
                        if not ok:
                            raise ValueError(
                                f"{self.name}: Stored idata entry '{mname}:{dname}' is turbine dependent, unable to reset n_turbines"
                            )
                        if FC.TURBINE in idata["coords"]:
                            idata["coords"][FC.TURBINE] = np.arange(self.n_turbines)
                        i = d[0].index(FC.TURBINE)
                        n0 = d[1].shape[i]
                        if n0 > self.n_turbines:
                            idata["data_vars"][dname] = (
                                d[0],
                                np.take(d[1], range(self.n_turbines), axis=i),
                            )
                        elif n0 < self.n_turbines:
                            shp = [
                                d[1].shape[j] if j != i else self.n_turbines - n0
                                for j in range(len(d[1].shape))
                            ]
                            a = np.zeros(shp, dtype=d[1].dtype)
                            shp = [
                                d[1].shape[j] if j != i else 1
                                for j in range(len(d[1].shape))
                            ]
                            a[:] = np.take(d[1], -1, axis=i).reshape(shp)
                            idata["data_vars"][dname] = (
                                d[0],
                                np.append(d[1], a, axis=i),
                            )

            self._idata_mem.update(newk)

    def get_models_idata(self):
        """
        Returns idata object of models

        Returns
        -------
        idata: dict, optional
            The dict has exactly two entries: `data_vars`,
            a dict with entries `name_str -> (dim_tuple, data_ndarray)`;
            and `coords`, a dict with entries `dim_name_str -> dim_array`.
            Take algorithm's idata object by default.

        """
        if not self.initialized:
            raise ValueError(
                f"Algorithm '{self.name}': get_models_idata called before initialization"
            )
        idata = {"coords": {}, "data_vars": {}}
        for k, hidata in self._idata_mem.items():
            if len(k) < 3 or k[:2] != "__":
                idata["coords"].update(hidata["coords"])
                idata["data_vars"].update(hidata["data_vars"])
        return idata

    def get_models_data(self, idata=None):
        """
        Creates xarray from model input data.

        Parameters
        ----------
        idata: dict, optional
            The dict has exactly two entries: `data_vars`,
            a dict with entries `name_str -> (dim_tuple, data_ndarray)`;
            and `coords`, a dict with entries `dim_name_str -> dim_array`.
            Take algorithm's idata object by default.

        Returns
        -------
        xarray.Dataset
            The model input data

        """
        if idata is None:
            idata = self.get_models_idata()
        sizes = self.__get_sizes(idata, "models")
        return self.__get_xrdata(idata, sizes)

    def new_point_data(self, points, states_indices=None):
        """
        Creates a point data xarray object, containing only points.

        Parameters
        ----------
        points: numpy.ndarray
            The points, shape: (n_states, n_points, 3)
        states_indices: array_like, optional
            The indices of the states dimension

        Returns
        -------
        xarray.Dataset
            A dataset containing the points data

        """

        if states_indices is None:
            idata = {"coords": {}, "data_vars": {}}
        else:
            idata = {"coords": {FC.STATE: states_indices}, "data_vars": {}}

        if (
            len(points.shape) != 3
            or points.shape[0] != self.n_states
            or points.shape[2] != 3
        ):
            raise ValueError(
                f"points have wrong dimensions, expecting ({self.n_states}, {points.shape[1]}, 3), got {points.shape}"
            )
        idata["data_vars"][FC.TARGETS] = (
            (FC.STATE, FC.TARGET, FC.TPOINT, FC.XYH),
            points[:, :, None, :],
        )
        idata["data_vars"][FC.TWEIGHTS] = (
            (FC.TPOINT,),
            np.array([1.0], dtype=FC.DTYPE),
        )

        sizes = self.__get_sizes(idata, "point")
        return self.__get_xrdata(idata, sizes)

    def finalize(self, clear_mem=False):
        """
        Finalizes the algorithm.

        Parameters
        ----------
        clear_mem: bool
            Clear idata memory

        """
        super().finalize(self, self.verbosity)
        if clear_mem:
            self._idata_mem = Dict()

    @classmethod
    def new(cls, algo_type, *args, **kwargs):
        """
        Run-time algorithm factory.

        Parameters
        ----------
        algo_type: str
            The selected derived class name
        args: tuple, optional
            Additional parameters for the constructor
        kwargs: dict, optional
            Additional parameters for the constructor

        """

        if algo_type is None:
            return None

        allc = all_subclasses(cls)
        found = algo_type in [scls.__name__ for scls in allc]

        if found:
            for scls in allc:
                if scls.__name__ == algo_type:
                    return scls(*args, **kwargs)

        else:
            estr = (
                "Algorithm type '{}' is not defined, available types are \n {}".format(
                    algo_type, sorted([i.__name__ for i in allc])
                )
            )
            raise KeyError(estr)



import numpy as np

import foxes.variables as FV
import foxes.constants as FC
from .rotor_points import RotorPoints


class PartialCentre(RotorPoints):
    """
    Partial wakes calculated only at the
    rotor centre point.

    :group: models.partial_wakes

    """

    def get_wake_points(self, algo, mdata, fdata):
        """
        Get the wake calculation points, and their
        weights.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        mdata: foxes.core.MData
            The model data
        fdata: foxes.core.FData
            The farm data

        Returns
        -------
        rpoints: numpy.ndarray
            The wake calculation points, shape:
            (n_states, n_turbines, n_tpoints, 3)
        rweights: numpy.ndarray
            The target point weights, shape: (n_tpoints,)

        """
        return fdata[FV.TXYH][:, :, None], np.ones(1, dtype=FC.DTYPE)



import numpy as np

from foxes.core import VerticalProfile
from foxes.utils import abl
import foxes.constants as FC
import foxes.variables as FV


class ABLLogWsProfile(VerticalProfile):
    """
    The neutral/stable/unstable ABL wind speed log profile.

    This profile picks the profile according to the mol value
    (neutral: mol = None or mol = 0)

    Attributes
    ----------
    ustar_input: bool
        Flag for using ustar as an input

    :group: models.vertical_profiles

    """

    def __init__(self, *args, ustar_input=False, **kwargs):
        """
        Constructor.

        Parameters
        ----------
        args: tuple, optional
            Additional arguments for VerticalProfile
        ustar_input: bool
            Flag for using ustar as an input
        kwargs: dict, optional
            Additional arguments for VerticalProfile

        """
        super().__init__(*args, **kwargs)
        self.ustar_input = ustar_input

    def input_vars(self):
        """
        The input variables needed for the profile
        calculation.

        Returns
        -------
        vars: list of str
            The variable names

        """
        if self.ustar_input:
            return [FV.USTAR, FV.Z0, FV.MOL]
        else:
            return [FV.WS, FV.H, FV.Z0, FV.MOL]

    def calculate(self, data, heights):
        """
        Run the profile calculation.

        Parameters
        ----------
        data: dict
            The input data
        heights: numpy.ndarray
            The evaluation heights

        Returns
        -------
        results: numpy.ndarray
            The profile results, same
            shape as heights

        """

        z0 = np.zeros_like(heights)
        z0[:] = data[FV.Z0]

        mol = np.zeros_like(heights)
        mol[:] = data[FV.MOL]

        if self.ustar_input:
            ustar = data[FV.USTAR]
        else:
            ws = np.zeros_like(heights)
            ws[:] = data[FV.WS]

            h0 = np.zeros_like(heights)
            h0[:] = data[FV.H]

        out = np.zeros_like(heights)

        # neutral profiles:
        sel = np.isnan(mol) | (mol == 0.0)
        if np.any(sel):
            sz0 = z0[sel]
            sh = heights[sel]
            if self.ustar_input:
                sus = ustar[sel]
            else:
                sus = abl.neutral.ustar(ws[sel], h0[sel], sz0, kappa=FC.KAPPA)
            out[sel] = abl.neutral.calc_ws(sh, sz0, sus, kappa=FC.KAPPA)

        # stable profiles:
        sel = mol > 0.0
        if np.any(sel):
            sz0 = z0[sel]
            smo = mol[sel]
            sh = heights[sel]
            if self.ustar_input:
                sus = ustar[sel]
            else:
                sus = abl.stable.ustar(ws[sel], h0[sel], sz0, smo, kappa=FC.KAPPA)
            psi = abl.stable.psi(sh, smo)
            out[sel] = abl.stable.calc_ws(sh, sz0, sus, psi, kappa=FC.KAPPA)

        # unstable profiles:
        sel = mol < 0.0
        if np.any(sel):
            sz0 = z0[sel]
            smo = mol[sel]
            sh = heights[sel]
            if self.ustar_input:
                sus = ustar[sel]
            else:
                sus = abl.unstable.ustar(ws[sel], h0[sel], sz0, smo, kappa=FC.KAPPA)
            psi = abl.unstable.psi(sh, smo)
            out[sel] = abl.unstable.calc_ws(sh, sz0, sus, psi, kappa=FC.KAPPA)

        return out



from pathlib import Path

from foxes.utils import PandasFileHelper, all_subclasses


class Output:
    """
    Base class for foxes output.

    The job of this class is to provide handy
    helper functions.

    Attributes
    ----------
    out_dir: pathlib.Path
        The output file directory
    out_fname_fun: Function, optional
        Modifies file names by f(fname)

    :group: output

    """

    def __init__(self, out_dir=None, out_fname_fun=None):
        """
        Constructor.

        Parameters
        ----------
        out_dir: str, optional
            The output file directory
        out_fname_fun: Function, optional
            Modifies file names by f(fname)

        """
        self.out_dir = Path(out_dir) if out_dir is not None else None
        self.out_fname_fun = out_fname_fun

    def get_fpath(self, fname):
        """
        Gets the total file path

        Parameters
        ----------
        fname: str
            The file name

        Returns
        -------
        fpath: pathlib.Path
            The total file path

        """
        fnm = Path(fname)
        if self.out_fname_fun is not None:
            fnm = self.out_fname_fun(fnm)
        return self.out_dir / fnm if self.out_dir is not None else fnm

    def write(self, file_name, data, format_col2var={}, format_dict={}, **kwargs):
        """
        Writes data to file via pandas.

        The kwargs are forwarded to the underlying pandas writing function.

        Parameters
        ----------
        file_name: str
            The output file name
        data: pandas.DataFrame
            The data
        format_col2var: dict
            Mapping from column names to foxes variables,
            for formatting
        format_dict: dict
            Dictionary with format entries for columns, e.g.
            {FV.P: '{:.4f}'}. Note that the keys are foxes variables

        """
        fdict = {}
        for c in data.columns:
            v = format_col2var.get(c, c)
            if v in format_dict:
                fdict[c] = format_dict[v]
            elif v in PandasFileHelper.DEFAULT_FORMAT_DICT:
                fdict[c] = PandasFileHelper.DEFAULT_FORMAT_DICT[v]

        fpath = self.get_fpath(file_name)
        PandasFileHelper.write_file(data, fpath, fdict, **kwargs)

    @classmethod
    def print_models(cls):
        """
        Prints all model names.
        """
        names = sorted([scls.__name__ for scls in all_subclasses(cls)])
        for n in names:
            print(n)

    @classmethod
    def new(cls, output_type, *args, **kwargs):
        """
        Run-time output model factory.

        Parameters
        ----------
        output_type: string
            The selected derived class name
        args: tuple, optional
            Additional parameters for the constructor
        kwargs: dict, optional
            Additional parameters for the constructor

        """

        if output_type is None:
            return None

        allc = all_subclasses(cls)
        found = output_type in [scls.__name__ for scls in allc]

        if found:
            for scls in allc:
                if scls.__name__ == output_type:
                    return scls(*args, **kwargs)

        else:
            estr = "Output type '{}' is not defined, available types are \n {}".format(
                output_type, sorted([i.__name__ for i in allc])
            )
            raise KeyError(estr)



import numpy as np

from foxes.opt.core.farm_opt_problem import FarmOptProblem
import foxes.variables as FV
import foxes.constants as FC


class FarmLayoutOptProblem(FarmOptProblem):
    """
    The turbine positioning optimization problem

    :group: opt.problems.layout

    """

    def var_names_float(self):
        """
        The names of float variables.

        Returns
        -------
        names: list of str
            The names of the float variables

        """
        vrs = []
        for ti in self.sel_turbines:
            vrs += [self.tvar(FV.X, ti), self.tvar(FV.Y, ti)]
        return vrs

    def initial_values_float(self):
        """
        The initial values of the float variables.

        Returns
        -------
        values: numpy.ndarray
            Initial float values, shape: (n_vars_float,)

        """
        out = np.zeros((self.n_sel_turbines, 2), dtype=FC.DTYPE)
        for i, ti in enumerate(self.sel_turbines):
            out[i] = self.farm.turbines[ti].xy
        return out.reshape(self.n_sel_turbines * 2)

    def min_values_float(self):
        """
        The minimal values of the float variables.

        Use -numpy.inf for unbounded.

        Returns
        -------
        values: numpy.ndarray
            Minimal float values, shape: (n_vars_float,)

        """
        b = self.farm.boundary
        assert b is not None, f"Problem '{self.name}': Missing wind farm boundary."
        out = np.zeros((self.n_sel_turbines, 2), dtype=FC.DTYPE)
        out[:] = b.p_min()[None, :]
        return out.reshape(self.n_sel_turbines * 2)

    def max_values_float(self):
        """
        The maximal values of the float variables.

        Use numpy.inf for unbounded.

        Returns
        -------
        values: numpy.ndarray
            Maximal float values, shape: (n_vars_float,)

        """
        b = self.farm.boundary
        assert b is not None, f"Problem '{self.name}': Missing wind farm boundary."
        out = np.zeros((self.n_sel_turbines, 2), dtype=FC.DTYPE)
        out[:] = b.p_max()[None, :]
        return out.reshape(self.n_sel_turbines * 2)

    def update_problem_individual(self, vars_int, vars_float):
        """
        Update the algo and other data using
        the latest optimization variables.

        This function is called before running the farm
        calculation.

        Parameters
        ----------
        vars_int: np.array
            The integer variable values, shape: (n_vars_int,)
        vars_float: np.array
            The float variable values, shape: (n_vars_float,)

        """
        super().update_problem_individual(vars_int, vars_float)

        xy = vars_float.reshape(self.n_sel_turbines, 2)
        for i, ti in enumerate(self.sel_turbines):
            t = self.algo.farm.turbines[ti]
            t.xy = xy[i]

    def update_problem_population(self, vars_int, vars_float):
        """
        Update the algo and other data using
        the latest optimization variables.

        This function is called before running the farm
        calculation.

        Parameters
        ----------
        vars_int: np.array
            The integer variable values, shape: (n_pop, n_vars_int,)
        vars_float: np.array
            The float variable values, shape: (n_pop, n_vars_float,)

        """
        super().update_problem_population(vars_int, vars_float)

        n_pop = len(vars_float)
        n_ostates = self._org_n_states
        n_states = n_pop * n_ostates

        xy = vars_float.reshape(n_pop, self.n_sel_turbines, 2)
        sxy = np.zeros(
            (n_pop, n_ostates, self.n_sel_turbines, 2), dtype=vars_float.dtype
        )
        sxy[:] = xy[:, None, :, :]
        sxy = sxy.reshape(n_states, self.n_sel_turbines, 2)
        del xy

        for i, ti in enumerate(self.sel_turbines):
            t = self.algo.farm.turbines[ti]
            t.xy = sxy[:, i]



def calc_ws(height, height0, WS0, shear):
    """
    Calculate wind speeds at given height

    Parameters
    ----------
    height: float
        The evaluation height
    height0: float
        Reference height
    WS0: float
        Reference wind speed
    shear: float
        Shear exponent

    Returns
    -------
    ws: float
        The wind speed

    :group: utils.abl.sheared

    """
    return WS0 * (height / height0) ** shear



import yaml

from foxes.core import Algorithm
from foxes.output import Output


def _write_yaml(data, fpath):
    """ Write the data to yaml """
    rmap = {
        "include": "!include",
    }
    with open(fpath, "w") as file:
        yaml.dump(data, file)
    with open(fpath, "r") as f:
        s = f.read()
    with open(fpath, 'w') as f:
        for k1, k2 in rmap.items():
            s = s.replace(k1, k2)
        f.write(s)
        
class WindioRunner:
    """
    Runner for windio input

    Attributes
    ----------
    algo: foxes.core.Algorithm
        The algorithm object
    output_dir: pathlib.Path
        Path to the output folder
    output_dicts: list of dict
        The output dictionaries
    farm_results: xarray.Dataset
        The farm results
    output_results: list
        The output results
    wio_input_data: dict
        The wind_energy_system windio input data
    file_name_input_yaml: str
        Name of the written input data file
    file_name_output_yaml: str
        Name of the written output data file
    verbosity: int
        The verbosity level, 0 = silent

    :group: input.windio

    """

    def __init__(
        self, 
        algo_dict, 
        output_dir=".",
        output_dicts=[], 
        wio_input_data=None, 
        file_name_input_yaml="recorded_input.yaml",
        file_name_output_yaml="recorded_output.yaml",
        verbosity=1,
        ):
        """
        Conbstructor

        Parameters
        ----------
        algo_dict: dict
            The algorithm dictionary
        output_dir: pathlib.Path
            Path to the output folder
        output_dicts: list of dict
            The output dictionaries
        wio_input_data: dict
            The wind_energy_system windio input data
        file_name_input_yaml: str
            Name of the written input data file
        file_name_output_yaml: str
            Name of the written output data file
        verbosity: int
            The verbosity level, 0 = silent

        """
        self.algo = algo_dict
        self.output_dir = output_dir
        self.output_dicts = output_dicts
        self.wio_input_data = wio_input_data
        self.file_name_input_yaml = file_name_input_yaml
        self.file_name_output_yaml = file_name_output_yaml
        self.verbosity = verbosity
        self.farm_results = None
        self.output_results = None

        self.__initialized = False
        
        self._output_yaml = {}
        if wio_input_data is not None and len(wio_input_data):
            fpath = output_dir/file_name_input_yaml
            self.print(f"Writing file", fpath)
            _write_yaml(wio_input_data, fpath)
            self._output_yaml["wind_energy_system"] = f"include {file_name_input_yaml}"

    def print(self, *args, level=1, **kwargs):
        """Print based on verbosity"""
        if self.verbosity >= level:
            print(*args, **kwargs)

    def initialize(self):
        """Initializes the runner"""
        if isinstance(self.algo, dict):
            self.print(f"Creating algorithm '{self.algo['algo_type']}'", level=2)
            self.algo = Algorithm.new(**self.algo)
        if not self.algo.initialized:
            self.algo.initialize()
        self.__initialized = True

    @property
    def initialized(self):
        """Flag for initialization"""
        return self.__initialized

    def run_farm_calc(self):
        """Runs the farm calculation"""
        if not self.__initialized:
            self.initialize()
        self.print("Running farm_calc")
        self.farm_results = self.algo.calc_farm()

    def run_outputs(self):
        """Runs the output calculation"""
        self.output_results = []
        for odict in self.output_dicts:
            self.print("Running output:", odict["output_type"])
            run_fname = odict.pop("run_func")
            run_args = odict.pop("run_args", ())
            run_kwargs = odict.pop("run_kwargs", {})
                    
            _odict = odict.copy()
            if "output_yaml_update" in _odict:
                self._output_yaml.update(_odict.pop("output_yaml_update"))
            if _odict.pop("farm_results", False):
                _odict["farm_results"] = self.farm_results
            if _odict.pop("algo", False):
                _odict["algo"] = self.algo
            o = Output.new(**_odict)
            f = getattr(o, run_fname)
            self.output_results.append(f(*run_args, **run_kwargs))
            
        fpath = self.output_dir/self.file_name_output_yaml
        self.print(f"Writing file", fpath)
        _write_yaml(self._output_yaml, fpath)
            
    def run(self):
        """Runs all calculations"""
        self.run_farm_calc()
        self.run_outputs()

    def finalize(self):
        """Initializes the runner"""
        if self.algo.initialized:
            self.algo.finalize(clear_mem=True)
        self.algo = None
        self.farm_results = None
        self.output_results = None
        self.__initialized = False

    def __enter__(self):
        self.initialize()
        return self

    def __exit__(self, *args):
        self.finalize()



from .from_csv import add_from_csv


def add_from_df(
    farm,
    data_source,
    *args,
    **kwargs,
):
    """
    Add turbines to wind farm via pandas DataFrame.

    Parameters
    ----------
    farm: foxes.WindFarm
        The wind farm
    data_source: str or pandas.DataFrame
        The input csv file or data source
    args: tuple, optional
        Additional parameters for add_from_csv()
    kwargs: dict, optional
        Additional parameters for add_from_csv()

    :group: input.farm_layout

    """
    add_from_csv(farm, data_source, *args, **kwargs)



import pandas as pd
from pathlib import Path
import xarray
from copy import deepcopy

import foxes.variables as FV


class PandasFileHelper:
    """
    This class helps reading and writing data
    to files via pandas.

    Attributes
    ----------
    DEFAULT_READING_PARAMETERS: dict
        Default parameters for file reading
        for the supported file formats
    DEFAULT_WRITING_PARAMETERS: dict
        Default parameters for file writing
        for the supported file formats
    DATA_FILE_FORMAT: list:str
        The supported file formats for data export
    DEFAULT_FORMAT_DICT: dict
        Default column formatting

    :group: utils

    """

    DEFAULT_READING_PARAMETERS = {
        "csv": {},
        "csv.gz": {},
        "csv.bz2": {},
        "csv.zip": {},
        "h5": {},
        "nc": {},
    }

    DEFAULT_WRITING_PARAMETERS = {
        "csv": {},
        "csv.gz": {},
        "csv.bz2": {},
        "csv.zip": {},
        "h5": {"key": "foxes", "mode": "w"},
        "nc": {},
    }

    DEFAULT_FORMAT_DICT = {
        FV.WD: "{:.3f}",
        FV.AMB_WD: "{:.3f}",
        FV.YAW: "{:.3f}",
        FV.AMB_YAW: "{:.3f}",
        FV.WS: "{:.4f}",
        FV.AMB_WS: "{:.4f}",
        FV.REWS: "{:.4f}",
        FV.AMB_REWS: "{:.4f}",
        FV.REWS2: "{:.4f}",
        FV.AMB_REWS2: "{:.4f}",
        FV.REWS3: "{:.4f}",
        FV.AMB_REWS3: "{:.4f}",
        FV.TI: "{:.6f}",
        FV.AMB_TI: "{:.6f}",
        FV.RHO: "{:.5f}",
        FV.AMB_RHO: "{:.5f}",
        FV.P: "{:.3f}",
        FV.AMB_P: "{:.3f}",
        FV.CT: "{:.6f}",
        FV.AMB_CT: "{:.6f}",
        FV.T: "{:.3f}",
        FV.AMB_T: "{:.3f}",
        FV.YLD: "{:.3f}",
        FV.AMB_YLD: "{:.3f}",
        FV.CAP: "{:.5f}",
        FV.AMB_CAP: "{:.5f}",
        FV.EFF: "{:.5f}",
    }

    DATA_FILE_FORMATS = list(DEFAULT_READING_PARAMETERS.keys())

    @classmethod
    def read_file(cls, file_path, **kwargs):
        """
        Helper for reading data according to file ending.

        Parameters
        ----------
        file_path: str
            The path to the file
        **kwargs: dict, optional
            Parameters forwarded to the pandas reading method.

        Returns
        -------
        pandas.DataFrame :
            The data

        """
        fpath = Path(file_path)
        fname = fpath.name
        sfx = ".".join(fname.split(".")[1:])
        f = None
        for fmt in cls.DATA_FILE_FORMATS:
            if sfx[:3] == "csv":
                f = pd.read_csv
            elif sfx == "h5":
                f = pd.read_hdf
            elif sfx == "nc":
                f = lambda fname, **pars: xarray.open_dataset(
                    fname, **pars
                ).to_dataframe()

            if f is not None:
                pars = deepcopy(cls.DEFAULT_READING_PARAMETERS[fmt])
                pars.update(kwargs)
                return f(file_path, **pars)

        raise KeyError(
            f"Unknown file format '{fname}'. Supported formats: {cls.DATA_FILE_FORMATS}"
        )

    @classmethod
    def write_file(cls, data, file_path, format_dict={}, **kwargs):
        """
        Helper for writing data according to file ending.

        Parameters
        ----------
        data: pandas.DataFrame
            The data
        file_path: str
            The path to the file
        format_dict: dict
            Dictionary with format entries for
            columns, e.g. '{:.4f}'
        **kwargs: dict, optional
            Parameters forwarded to the pandas writing method.

        """

        fdict = deepcopy(cls.DEFAULT_FORMAT_DICT)
        fdict.update(format_dict)

        out = pd.DataFrame(index=data.index, columns=data.columns)
        for c in data.columns:
            if c in fdict.keys():
                out[c] = data[c].map(
                    lambda x: fdict[c].format(x) if not pd.isna(x) else x
                )
            else:
                out[c] = data[c]

        fpath = Path(file_path)
        fname = fpath.name
        sfx = ".".join(fname.split(".")[1:])
        f = None
        for fmt in cls.DATA_FILE_FORMATS:
            if sfx[:3] == "csv":
                f = out.to_csv
            elif sfx == "h5":
                f = out.to_hdf
            elif sfx == "nc":
                f = out.to_netcdf

            if f is not None:
                pars = cls.DEFAULT_WRITING_PARAMETERS[fmt]
                pars.update(kwargs)

                f(file_path, **pars)

                return

        raise KeyError(
            f"Unknown file format '{file_path}'. Supported formats: {cls.DATA_FILE_FORMATS}"
        )



import numpy as np

from foxes.models.wake_models.top_hat import TopHatWakeModel
from foxes.utils.two_circles import calc_area
import foxes.variables as FV
import foxes.constants as FC

from .centre import PartialCentre


class PartialTopHat(PartialCentre):
    """
    Partial wakes for top-hat models.

    The wake effect is weighted by the overlap of
    the wake circle and the rotor disc circle.

    Attributes
    ----------
    rotor_model: foxes.core.RotorModel
        The rotor model, default is the one from the algorithm

    :group: models.partial_wakes

    """

    def check_wmodel(self, wmodel, error=True):
        """
        Checks the wake model type

        Parameters
        ----------
        wmodel: foxes.core.WakeModel
            The wake model to be tested
        error: bool
            Flag for raising TypeError

        Returns
        -------
        chk: bool
            True if wake model is compatible

        """
        if not isinstance(wmodel, TopHatWakeModel):
            if error:
                raise TypeError(
                    f"Partial wakes '{self.name}': Cannot be applied to wake model '{wmodel.name}', since not a TopHatWakeModel"
                )
            return False
        return True

    def __init__(self, rotor_model=None):
        """
        Constructor.

        Parameters
        ----------
        rotor_model: foxes.core.RotorModel, optional
            The rotor model, default is the one from the algorithm

        """
        super().__init__()
        self.rotor_model = rotor_model

    def initialize(self, algo, verbosity=0):
        """
        Initializes the model.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        verbosity: int
            The verbosity level, 0 = silent

        """
        if self.rotor_model is None:
            self.rotor_model = algo.rotor_model

        super().initialize(algo, verbosity)

        self.WCOOS_ID = self.var("WCOOS_ID")
        self.WCOOS_X = self.var("WCOOS_X")
        self.WCOOS_R = self.var("WCOOS_R")

    def sub_models(self):
        """
        List of all sub-models

        Returns
        -------
        smdls: list of foxes.core.Model
            Names of all sub models

        """
        return super().sub_models() + [self.rotor_model]

    def contribute(
        self,
        algo,
        mdata,
        fdata,
        tdata,
        downwind_index,
        wake_deltas,
        wmodel,
    ):
        """
        Modifies wake deltas at target points by
        contributions from the specified wake source turbines.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        mdata: foxes.core.MData
            The model data
        fdata: foxes.core.FData
            The farm data
        tdata: foxes.core.TData
            The target point data
        downwind_index: int
            The index of the wake causing turbine
            in the downwnd order
        wake_deltas: dict
            The wake deltas. Key: variable name,
            value: numpy.ndarray with shape
            (n_states, n_targets, n_tpoints, ...)
        wmodel: foxes.core.WakeModel
            The wake model

        """
        self.check_wmodel(wmodel, error=True)

        wcoos = algo.wake_frame.get_wake_coos(algo, mdata, fdata, tdata, downwind_index)
        x = wcoos[:, :, 0, 0]
        yz = wcoos[:, :, 0, 1:3]
        del wcoos

        ct = self.get_data(
            FV.CT,
            FC.STATE_TARGET,
            lookup="w",
            fdata=fdata,
            tdata=tdata,
            downwind_index=downwind_index,
            algo=algo,
            upcast=True,
        )

        sel0 = (ct > 0) & (x > 0)
        if np.any(sel0):
            R = np.linalg.norm(yz, axis=-1)
            del yz

            D = self.get_data(
                FV.D,
                FC.STATE_TARGET,
                lookup="w",
                fdata=fdata,
                tdata=tdata,
                downwind_index=downwind_index,
                algo=algo,
                upcast=True,
            )

            wr = wmodel.calc_wake_radius(
                algo, mdata, fdata, tdata, downwind_index, x, ct
            )

            st_sel = sel0 & (wr > R - D / 2)
            if np.any(st_sel):
                x = x[st_sel]
                ct = ct[st_sel]
                wr = wr[st_sel]
                R = R[st_sel]
                D = D[st_sel]

                clw = wmodel.calc_centreline(
                    algo, mdata, fdata, tdata, downwind_index, st_sel, x, wr, ct
                )

                weights = calc_area(D / 2, wr, R) / (np.pi * (D / 2) ** 2)

                for v, d in clw.items():
                    try:
                        superp = wmodel.superp[v]
                    except KeyError:
                        raise KeyError(
                            f"Model '{self.name}': Missing wake superposition entry for variable '{v}' in wake model '{wmodel.name}', found {sorted(list(wmodel.superp.keys()))}"
                        )

                    wake_deltas[v] = superp.add_wake(
                        algo,
                        mdata,
                        fdata,
                        tdata,
                        downwind_index,
                        st_sel,
                        v,
                        wake_deltas[v],
                        weights[:, None] * d[:, None],
                    )



import matplotlib.pyplot as plt
import numpy as np
from mpl_toolkits.axes_grid1 import make_axes_locatable

from foxes.utils import wd2uv


def get_fig(
    var,
    fig,
    figsize,
    ax,
    data,
    si,
    s,
    levels,
    x_pos,
    y_pos,
    cmap,
    xlabel,
    ylabel,
    title,
    add_bar,
    vlabel,
    ret_state,
    ret_im,
    vmin=None,
    vmax=None,
    quiv=None,
    invert_axis=None,
    animated=False,
    show_rotor_dict=None,
):
    """
    Helper function that creates the flow image plot.

    Parameters
    ----------
    var: str
        The variable name
    fig: plt.Figure, optional
        The figure object
    figsize: tuple
        The figsize for plt.Figure
    ax: plt.Axes, optional
        The figure axes
    data: numpy.ndarray
        The grid data to plot, shape: (n_states, n_x, x_y)
    si: int, optional
        The state counter
    s: object
        The state index
    levels: int
        The number of levels for the contourf plot,
        or None for non-contour image
    x_pos: numpy.ndarray
        The grid x positions, shape: (n_x, 3)
    y_pos: numpy.ndarray
        The grid y positions, shape: (n_y, 3)
    xlabel: str
        The x axis label
    ylabel: str
        The y axis label
    title: str, optional
        The title
    add_bar: bool
        Add a color bar
    vlabel: str, optional
        The variable label
    ret_state: bool
        Flag for state index return
    ret_im: bool
        Flag for image return
    vmin: float, optional
        The minimal variable value
    vmax: float, optional
        The maximal variable value
    quiv: tuple, optional
        The quiver data: (n, pars, wd, ws)
    invert_axis: str, optional
        Which axis to invert, either x or y
    animated: bool
        Switch for usage for an animation
    show_rotor_dict: dict, optional
        Parameters for indicating the rotor plane
        by a line

    Yields
    ------
    fig: matplotlib.Figure
        The figure object
    si: int, optional
        The state index
    im: tuple, optional
        The image objects, matplotlib.collections.QuadMesh
        or matplotlib.QuadContourSet

    """
    # create plot:
    if fig is None:
        hfig = plt.figure(figsize=figsize)
    else:
        hfig = fig
    if ax is None:
        hax = hfig.add_subplot(111)
    else:
        hax = ax

    # get results:
    N_x = len(x_pos)
    N_y = len(y_pos)
    zz = data[si].reshape(N_x, N_y).T

    # raw data image:
    if levels is None:
        im = hax.pcolormesh(
            x_pos,
            y_pos,
            zz,
            shading="auto",
            cmap=cmap,
            vmin=vmin,
            vmax=vmax,
            animated=animated,
        )

    # contour plot:
    else:
        im = hax.contourf(
            x_pos,
            y_pos,
            zz,
            levels=levels,
            cmap=cmap,
            vmin=vmin,
            vmax=vmax,
            # animated=animated,
        )

    qv = None
    if quiv is not None and quiv[0] is not None:
        n, pars, wd, ws = quiv
        uv = wd2uv(wd[si], ws[si])
        u = uv[:, :, 0].T[::n, ::n]
        v = uv[:, :, 1].T[::n, ::n]
        qv = hax.quiver(x_pos[::n], y_pos[::n], u, v, animated=animated, **pars)
        del n, pars, u, v, uv

    hax.autoscale_view()
    hax.set_xlabel(xlabel)
    hax.set_ylabel(ylabel)
    hax.set_aspect("equal", adjustable="box")
    hax.set_xlim(x_pos.min(), x_pos.max())
    hax.set_ylim(y_pos.min(), y_pos.max())

    ttl = None
    if animated:
        if title is None:
            if hasattr(s, "dtype") and np.issubdtype(s.dtype, np.datetime64):
                t = np.datetime_as_string(s, unit="m").replace("T", " ")
            else:
                t = f"State {s}"
        else:
            t = title
        ttl = hax.text(
            0.5,
            1.05,
            t,
            backgroundcolor="w",
            transform=hax.transAxes,
            ha="center",
            animated=True,
            clip_on=False,
        )
    else:
        hax.set_title(title if title is not None else f"State {s}")

    if invert_axis == "x":
        hax.invert_xaxis()
    elif invert_axis == "y":
        hax.invert_yaxis()

    # add rotor position:
    if show_rotor_dict is not None:
        D = show_rotor_dict["D"]
        coords = np.zeros(shape=(2, len(D)))  # array to hold change to turbine coords

        if (xlabel == "x [m]") & (ylabel == "y [m]"):
            x = show_rotor_dict["X"]
            y = show_rotor_dict["Y"]
            turb_angle = show_rotor_dict["turb_angle"]
            theta = np.deg2rad(np.mod(turb_angle + 90, 360))
            coords[0, :] = (D / 2) * np.sin(theta)
            coords[1, :] = (D / 2) * np.cos(theta)

        if (xlabel == "x [m]") & (ylabel == "z [m]"):
            # get hub heights for use as y coords in plot
            x = show_rotor_dict["X"]
            y = show_rotor_dict["H"]
            coords[1, :] = D / 2  # don't show yaw in xz plot

        if (xlabel == "y [m]") & (ylabel == "z [m]"):
            x = show_rotor_dict["Y"]
            y = show_rotor_dict["H"]
            coords[1, :] = D / 2

        # plot each rotor
        c = show_rotor_dict["color"]
        for t in np.arange(len(D)):
            turb_x1 = x[t] + coords[0, t]
            turb_x2 = x[t] - coords[0, t]
            turb_y1 = y[t] + coords[1, t]
            turb_y2 = y[t] - coords[1, t]
            hax.plot(
                [turb_x1, turb_x2],
                [turb_y1, turb_y2],
                color=c,
                linestyle="-",
                linewidth=1,
            )

    if add_bar:
        divider = make_axes_locatable(hax)
        cax = divider.append_axes("right", size="5%", pad=0.05)
        vlab = vlabel if vlabel is not None else var
        hfig.colorbar(im, cax=cax, orientation="vertical", label=vlab)
        out = hfig
    else:
        out = fig

    if ret_state or ret_im:
        out = [out]
    if ret_state:
        out.append(si)
    if ret_im:
        out.append([i for i in [im, qv, ttl] if i is not None])
    if ret_state or ret_im:
        out = tuple(out)

    return out



import numpy as np

from foxes.core import WakeSuperposition
import foxes.variables as FV


class TIQuadratic(WakeSuperposition):
    """
    Quadratic wake superposition for TI.

    Attributes
    ----------
    superp_to_amb: str
        The method for combining ambient with wake deltas:
        linear or quadratic

    :group: models.wake_superpositions

    """

    def __init__(self, superp_to_amb="quadratic"):
        """
        Constructor.

        Parameters
        ----------
        superp_to_amb: str
            The method for combining ambient with wake deltas:
            linear or quadratic

        """
        super().__init__()
        self.superp_to_amb = superp_to_amb

    def __repr__(self):
        return f"{type(self).__name__}(superp_to_amb={self.superp_to_amb})"

    def add_wake(
        self,
        algo,
        mdata,
        fdata,
        tdata,
        downwind_index,
        st_sel,
        variable,
        wake_delta,
        wake_model_result,
    ):
        """
        Add a wake delta to previous wake deltas,
        at rotor points.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        mdata: foxes.core.MData
            The model data
        fdata: foxes.core.FData
            The farm data
        tdata: foxes.core.TData
            The target point data
        downwind_index: int
            The index of the wake causing turbine
            in the downwnd order
        st_sel: numpy.ndarray of bool
            The selection of targets, shape: (n_states, n_targets)
        variable: str
            The variable name for which the wake deltas applies
        wake_delta: numpy.ndarray
            The original wake deltas, shape:
            (n_states, n_targets, n_tpoints, ...)
        wake_model_result: numpy.ndarray
            The new wake deltas of the selected rotors,
            shape: (n_st_sel, n_tpoints, ...)

        Returns
        -------
        wdelta: numpy.ndarray
            The updated wake deltas, shape:
            (n_states, n_targets, n_tpoints, ...)

        """
        if variable != FV.TI:
            raise ValueError(
                f"Superposition '{self.name}': Expecting wake variable {FV.TI}, got {variable}"
            )

        wake_delta[st_sel] += wake_model_result**2
        return wake_delta

    def calc_final_wake_delta(
        self,
        algo,
        mdata,
        fdata,
        variable,
        amb_results,
        wake_delta,
    ):
        """
        Calculate the final wake delta after adding all
        contributions.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        mdata: foxes.core.MData
            The model data
        fdata: foxes.core.FData
            The farm data
        variable: str
            The variable name for which the wake deltas applies
        amb_results: numpy.ndarray
            The ambient results at targets,
            shape: (n_states, n_targets, n_tpoints)
        wake_delta: numpy.ndarray
            The wake deltas at targets, shape:
            (n_states, n_targets, n_tpoints)

        Returns
        -------
        final_wake_delta: numpy.ndarray
            The final wake delta, which will be added to the ambient
            results by simple plus operation. Shape:
            (n_states, n_targets, n_tpoints)

        """
        # linear superposition to ambient:
        if self.superp_to_amb == "linear":
            return np.sqrt(wake_delta)

        # quadratic superposition to ambient:
        elif self.superp_to_amb == "quadratic":
            return np.sqrt(wake_delta + amb_results**2) - amb_results

        # unknown ti delta:
        else:
            raise ValueError(
                f"Unknown superp_to_amb = '{self.superp_to_amb}', valid choices: linear, quadratic"
            )



import numpy as np
import pandas as pd
from xarray import Dataset
from pathlib import Path

from foxes.core import States, VerticalProfile
from foxes.utils import PandasFileHelper, read_tab_file
from foxes.data import STATES
import foxes.variables as FV
import foxes.constants as FC


class StatesTable(States):
    """
    States from a `pandas.DataFrame` or a pandas readable file.

    Attributes
    ----------
    data_source: str or pandas.DataFrame
        Either path to a file or data
    ovars: list of str
        The output variables
    var2col: dict
        Mapping from variable names to data column names
    fixed_vars: dict
        Fixed uniform variable values, instead of
        reading from data
    profdicts: dict
        Key: output variable name str, Value: str or dict
        or `foxes.core.VerticalProfile`
    rpars: dict
        pandas file reading parameters
    states_sel: slice or range or list of int
        States subset selection
    states_loc: list
        State index selection via pandas loc function
    RDICT: dict
        Default pandas file reading parameters

    :group: input.states

    """

    RDICT = {"index_col": 0}

    def __init__(
        self,
        data_source,
        output_vars,
        var2col={},
        fixed_vars={},
        profiles={},
        pd_read_pars={},
        states_sel=None,
        states_loc=None,
    ):
        """
        Constructor.

        Parameters
        ----------
        data_source: str or pandas.DataFrame
            Either path to a file or data
        output_vars: list of str
            The output variables
        var2col: dict
            Mapping from variable names to data column names
        fixed_vars: dict
            Fixed uniform variable values, instead of
            reading from data
        profiles: dict
            Key: output variable name str, Value: str or dict
            or `foxes.core.VerticalProfile`
        pd_read_pars: dict
            pandas file reading parameters
        states_sel: slice or range or list of int, optional
            States subset selection
        states_loc: list, optional
            State index selection via pandas loc function

        """
        super().__init__()

        self.data_source = data_source
        self.ovars = output_vars
        self.rpars = pd_read_pars
        self.var2col = var2col
        self.fixed_vars = fixed_vars
        self.profdicts = profiles
        self.states_sel = states_sel
        self.states_loc = states_loc

        if self.states_loc is not None and self.states_sel is not None:
            raise ValueError(
                f"States '{self.name}': Cannot handle both 'states_sel' and 'states_loc', please pick one"
            )

        self._weights = None
        self._N = None
        self._tvars = None
        self._profiles = None

    def reset(self, algo=None, states_sel=None, states_loc=None, verbosity=0):
        """
        Reset the states, optionally select states

        Parameters
        ----------
        states_sel: slice or range or list of int, optional
            States subset selection
        states_loc: list, optional
            State index selection via pandas loc function
        verbosity: int
            The verbosity level, 0 = silent

        """
        if self.initialized:
            if algo is None:
                raise KeyError(f"{self.name}: Missing algo for reset")
            elif algo.states is not self:
                raise ValueError(f"{self.states}: algo.states differs from self")
            self.finalize(algo, verbosity)
        self.states_sel = states_sel
        self.states_loc = states_loc

    def initialize(self, algo, verbosity=0):
        """
        Initializes the model.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        verbosity: int
            The verbosity level, 0 = silent

        """
        self._profiles = {}
        self._tvars = set(self.ovars)
        for v, d in self.profdicts.items():
            if isinstance(d, str):
                self._profiles[v] = VerticalProfile.new(d)
            elif isinstance(d, VerticalProfile):
                self._profiles[v] = d
            elif isinstance(d, dict):
                t = d.pop("type")
                self._profiles[v] = VerticalProfile.new(t, **d)
            else:
                raise TypeError(
                    f"States '{self.name}': Wrong profile type '{type(d).__name__}' for variable '{v}'. Expecting VerticalProfile, str or dict"
                )
            self._tvars.update(self._profiles[v].input_vars())
        self._tvars -= set(self.fixed_vars.keys())
        self._tvars = list(self._tvars)
        super().initialize(algo, verbosity)

    def sub_models(self):
        """
        List of all sub-models

        Returns
        -------
        smdls: list of str
            Names of all sub models

        """

        return list(self._profiles.values())

    def load_data(self, algo, verbosity=0):
        """
        Load and/or create all model data that is subject to chunking.

        Such data should not be stored under self, for memory reasons. The
        data returned here will automatically be chunked and then provided
        as part of the mdata object during calculations.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        verbosity: int
            The verbosity level, 0 = silent

        Returns
        -------
        idata: dict
            The dict has exactly two entries: `data_vars`,
            a dict with entries `name_str -> (dim_tuple, data_ndarray)`;
            and `coords`, a dict with entries `dim_name_str -> dim_array`

        """
        self.VARS = self.var("vars")
        self.DATA = self.var("data")

        if isinstance(self.data_source, pd.DataFrame):
            data = self.data_source
            isorg = True
        else:
            if not Path(self.data_source).is_file():
                if verbosity:
                    print(
                        f"States '{self.name}': Reading static data '{self.data_source}' from context '{STATES}'"
                    )
                self.data_source = algo.dbook.get_file_path(
                    STATES, self.data_source, check_raw=False
                )
                if verbosity:
                    print(f"Path: {self.data_source}")
            elif verbosity:
                print(f"States '{self.name}': Reading file {self.data_source}")
            rpars = dict(self.RDICT, **self.rpars)
            data = PandasFileHelper().read_file(self.data_source, **rpars)
            isorg = False

        if self.states_sel is not None:
            data = data.iloc[self.states_sel]
        elif self.states_loc is not None:
            data = data.loc[self.states_loc]
        self._N = len(data.index)
        self._inds = data.index.to_numpy()

        col_w = self.var2col.get(FV.WEIGHT, FV.WEIGHT)
        self._weights = np.zeros((self._N, algo.n_turbines), dtype=FC.DTYPE)
        if col_w in data:
            self._weights[:] = data[col_w].to_numpy()[:, None]
        elif FV.WEIGHT in self.var2col:
            raise KeyError(
                f"Weight variable '{col_w}' defined in var2col, but not found in states table columns {data.columns}"
            )
        else:
            self._weights[:] = 1.0 / self._N
            if isorg:
                data = data.copy()
            data[col_w] = self._weights[:, 0]

        tcols = []
        for v in self._tvars:
            c = self.var2col.get(v, v)
            if c in data.columns:
                tcols.append(c)
            elif v not in self._profiles.keys():
                raise KeyError(
                    f"States '{self.name}': Missing variable '{c}' in states table columns, profiles or fixed vars"
                )
        data = data[tcols]

        idata = super().load_data(algo, verbosity)
        idata["coords"][self.VARS] = self._tvars
        idata["data_vars"][self.DATA] = ((FC.STATE, self.VARS), data.to_numpy())

        return idata

    def size(self):
        """
        The total number of states.

        Returns
        -------
        int:
            The total number of states

        """
        return self._N

    def index(self):
        """
        The index list

        Returns
        -------
        indices: array_like
            The index labels of states, or None for default integers

        """
        return self._inds

    def output_point_vars(self, algo):
        """
        The variables which are being modified by the model.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm

        Returns
        -------
        output_vars: list of str
            The output variable names

        """
        return self.ovars

    def weights(self, algo):
        """
        The statistical weights of all states.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm

        Returns
        -------
        weights: numpy.ndarray
            The weights, shape: (n_states, n_turbines)

        """
        return self._weights

    def calculate(self, algo, mdata, fdata, tdata):
        """ "
        The main model calculation.

        This function is executed on a single chunk of data,
        all computations should be based on numpy arrays.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        mdata: foxes.core.MData
            The model data
        fdata: foxes.core.FData
            The farm data
        tdata: foxes.core.TData
            The target point data

        Returns
        -------
        results: dict
            The resulting data, keys: output variable str.
            Values: numpy.ndarray with shape
            (n_states, n_targets, n_tpoints)

        """
        for i, v in enumerate(self._tvars):
            if v in tdata:
                tdata[v][:] = mdata[self.DATA][:, i, None, None]
            else:
                tdata[v] = np.zeros(
                    (tdata.n_states, tdata.n_targets, tdata.n_tpoints), dtype=FC.DTYPE
                )
                tdata[v][:] = mdata[self.DATA][:, i, None, None]
                tdata.dims[v] = (FC.STATE, FC.TARGET, FC.TPOINT)

        for v, f in self.fixed_vars.items():
            tdata[v] = np.full(
                (tdata.n_states, tdata.n_targets, tdata.n_tpoints), f, dtype=FC.DTYPE
            )

        z = tdata[FC.TARGETS][..., 2]
        for v, p in self._profiles.items():
            tdata[v] = p.calculate(tdata, z)

        return {v: tdata[v] for v in self.output_point_vars(algo)}

    def finalize(self, algo, verbosity=0):
        """
        Finalizes the model.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        verbosity: int
            The verbosity level

        """
        self._weights = None
        self._N = None
        self._tvars = None

        super().finalize(algo, verbosity)


class Timeseries(StatesTable):
    """
    Timeseries states data.

    :group: input.states

    """

    RDICT = {"index_col": 0, "parse_dates": [0]}


class TabStates(StatesTable):
    """
    States created from a single tab file

    :group: input.states

    """

    def __init__(self, data_source, *args, normalize=True, **kwargs):
        """
        Constructor.

        Parameters
        ----------
        data_source: str or xarray.Dataset
            The tab file data file name, or its data
        args: tuple, optional
            Additional parameters for StatesTable
        normalize: bool
            Normalize the tab file data
        kwargs: dict, optional
            Additional parameters for StatesTable

        """
        self._normalize = normalize
        if isinstance(data_source, Dataset):
            self._tab_source = None
            self._tab_data = data_source
        elif isinstance(data_source, (str, Path)):
            self._tab_source = data_source
            self._tab_data = None
        else:
            raise TypeError(
                f"Expecting str, Path or xarray.Dataset as data_source, got {type(data_source)}"
            )

        super().__init__(data_source=None, *args, **kwargs)

    def load_data(self, algo, verbosity=0):
        """
        Load and/or create all model data that is subject to chunking.

        Such data should not be stored under self, for memory reasons. The
        data returned here will automatically be chunked and then provided
        as part of the mdata object during calculations.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        verbosity: int
            The verbosity level, 0 = silent

        Returns
        -------
        idata: dict
            The dict has exactly two entries: `data_vars`,
            a dict with entries `name_str -> (dim_tuple, data_ndarray)`;
            and `coords`, a dict with entries `dim_name_str -> dim_array`

        """
        if self.data_source is None:
            if self._tab_data is None:
                if not Path(self._tab_source).is_file():
                    if verbosity:
                        print(
                            f"States '{self.name}': Reading static data '{self._tab_source}' from context '{STATES}'"
                        )
                    self._tab_source = algo.dbook.get_file_path(
                        STATES, self._tab_source, check_raw=False
                    )
                    if verbosity:
                        print(f"Path: {self._tab_source}")
                elif verbosity:
                    print(f"States '{self.name}': Reading file {self._tab_source}")
                self._tab_data = read_tab_file(self._tab_source, self._normalize)

            a = self._tab_data.attrs["factor_ws"]
            b = self._tab_data.attrs["shift_wd"]
            if b != 0.0:
                raise ValueError(
                    f"{self.name}: shift_wd = {b} is not supported, expecting zero"
                )

            wd0 = self._tab_data["wd"].to_numpy()
            ws0 = a * np.append(
                np.array([0], dtype=FC.DTYPE), self._tab_data["ws"].to_numpy()
            )
            ws0 = 0.5 * (ws0[:-1] + ws0[1:])

            n_ws = self._tab_data.sizes["ws"]
            n_wd = self._tab_data.sizes["wd"]
            ws = np.zeros((n_ws, n_wd), dtype=FC.DTYPE)
            wd = np.zeros((n_ws, n_wd), dtype=FC.DTYPE)
            ws[:] = ws0[:, None]
            wd[:] = wd0[None, :]

            wd_freq = self._tab_data["wd_freq"].to_numpy() / 100
            weights = self._tab_data["ws_freq"].to_numpy() * wd_freq[None, :] / 1000

            sel = weights > 0

            self.data_source = pd.DataFrame(
                index=np.arange(np.sum(sel)),
                data={
                    FV.WS: ws[sel],
                    FV.WD: wd[sel],
                    FV.WEIGHT: weights[sel],
                },
            )
            self.data_source.index.name = FC.STATE

        return super().load_data(algo, verbosity)



import numpy as np

from foxes.core import TurbineInductionModel
from foxes.utils import uv2wd, wd2uv
import foxes.variables as FV
import foxes.constants as FC


class RankineHalfBody(TurbineInductionModel):
    """
    The Rankine half body induction wake model

    The individual wake effects are superposed linearly,
    without invoking a wake superposition model.

    Notes
    -----
    Reference:
    B Gribben and G Hawkes
    "A potential flow model for wind turbine induction and wind farm blockage"
    Techincal Paper, Frazer-Nash Consultancy, 2019
    https://www.fnc.co.uk/media/o5eosxas/a-potential-flow-model-for-wind-turbine-induction-and-wind-farm-blockage.pdf

    Attributes
    ----------
    induction: foxes.core.AxialInductionModel or str
        The induction model

    :group: models.wake_models.induction

    """

    def __init__(self, induction="Madsen"):
        """
        Constructor.

        Parameters
        ----------
        induction: foxes.core.AxialInductionModel or str
            The induction model

        """
        super().__init__()
        self.induction = induction

    def __repr__(self):
        iname = (
            self.induction if isinstance(self.induction, str) else self.induction.name
        )
        return f"{type(self).__name__}(induction={iname})"

    def sub_models(self):
        """
        List of all sub-models

        Returns
        -------
        smdls: list of foxes.core.Model
            All sub models

        """
        return [self.induction]

    def initialize(self, algo, verbosity=0, force=False):
        """
        Initializes the model.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        verbosity: int
            The verbosity level, 0 = silent
        force: bool
            Overwrite existing data

        """
        if isinstance(self.induction, str):
            self.induction = algo.mbook.axial_induction[self.induction]
        super().initialize(algo, verbosity, force)

    def new_wake_deltas(self, algo, mdata, fdata, tdata):
        """
        Creates new empty wake delta arrays.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        mdata: foxes.core.MData
            The model data
        fdata: foxes.core.FData
            The farm data
        tdata: foxes.core.TData
            The target point data

        Returns
        -------
        wake_deltas: dict
            Key: variable name, value: The zero filled
            wake deltas, shape: (n_states, n_turbines, n_rpoints, ...)

        """
        return {
            FV.WS: np.zeros_like(tdata[FC.TARGETS][..., 0]),
            FV.WD: np.zeros_like(tdata[FC.TARGETS][..., 0]),
            "U": np.zeros_like(tdata[FC.TARGETS][..., 0]),
            "V": np.zeros_like(tdata[FC.TARGETS][..., 0]),
        }

    def contribute(
        self,
        algo,
        mdata,
        fdata,
        tdata,
        downwind_index,
        wake_coos,
        wake_deltas,
    ):
        """
        Modifies wake deltas at target points by
        contributions from the specified wake source turbines.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        mdata: foxes.core.MData
            The model data
        fdata: foxes.core.FData
            The farm data
        tdata: foxes.core.TData
            The target point data
        downwind_index: int
            The index of the wake causing turbine
            in the downwnd order
        wake_coos: numpy.ndarray
            The wake frame coordinates of the evaluation
            points, shape: (n_states, n_targets, n_tpoints, 3)
        wake_deltas: dict
            The wake deltas. Key: variable name,
            value: numpy.ndarray with shape
            (n_states, n_targets, n_tpoints, ...)

        """
        # get ct:
        ct = self.get_data(
            FV.CT,
            FC.STATE_TARGET_TPOINT,
            lookup="w",
            algo=algo,
            fdata=fdata,
            tdata=tdata,
            downwind_index=downwind_index,
            upcast=False,
        )

        # get ws:
        ws = self.get_data(
            FV.REWS,
            FC.STATE_TARGET_TPOINT,
            lookup="w",
            algo=algo,
            fdata=fdata,
            tdata=tdata,
            downwind_index=downwind_index,
            upcast=False,
        )

        # get D
        D = self.get_data(
            FV.D,
            FC.STATE_TARGET_TPOINT,
            lookup="w",
            algo=algo,
            fdata=fdata,
            tdata=tdata,
            downwind_index=downwind_index,
            upcast=True,
        )

        # calc m (page 7, skipping pi everywhere)
        m = 2 * ws * self.induction.ct2a(ct) * (D / 2) ** 2

        # get r and theta
        x = np.round(wake_coos[..., 0], 12)
        r = np.linalg.norm(wake_coos[..., 1:], axis=-1)
        r_sph = np.sqrt(r**2 + x**2)
        theta = np.arctan2(r, x)

        # define rankine half body shape (page 3)
        RHB_shape = (
            np.cos(theta) - (2 / (m + 1e-15)) * ws * (r_sph * np.sin(theta)) ** 2
        )

        # stagnation point condition
        xs = -np.sqrt(m / (4 * ws + 1e-15))

        # set values out of body shape
        st_sel = (ct > 0) & ((RHB_shape < -1) | (x < xs))
        if np.any(st_sel):
            # apply selection
            xyz = wake_coos[st_sel]

            # calc velocity components
            vel_factor = m[st_sel] / (4 * np.linalg.norm(xyz, axis=-1) ** 3)
            wake_deltas["U"][st_sel] += vel_factor * xyz[:, 0]
            wake_deltas["V"][st_sel] += vel_factor * xyz[:, 1]

        # set values inside body shape
        st_sel = (ct > 0) & (RHB_shape >= -1) & (x >= xs) & (x <= 0)
        if np.any(st_sel):
            # apply selection
            xyz = np.zeros_like(wake_coos[st_sel])
            xyz[:, 0] = xs[st_sel]

            # calc velocity components
            vel_factor = m[st_sel] / (4 * np.linalg.norm(xyz, axis=-1) ** 3)
            wake_deltas["U"][st_sel] += vel_factor * xyz[:, 0]

    def finalize_wake_deltas(
        self,
        algo,
        mdata,
        fdata,
        amb_results,
        wake_deltas,
    ):
        """
        Finalize the wake calculation.

        Modifies wake_deltas on the fly.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        mdata: foxes.core.MData
            The model data
        fdata: foxes.core.FData
            The farm data
        amb_results: dict
            The ambient results, key: variable name str,
            values: numpy.ndarray with shape
            (n_states, n_targets, n_tpoints)
        wake_deltas: dict
            The wake deltas object at the selected target
            turbines. Key: variable str, value: numpy.ndarray
            with shape (n_states, n_targets, n_tpoints)

        """
        # calc ambient wind vector:
        ws0 = amb_results[FV.WS]
        nx = wd2uv(amb_results[FV.WD])
        wind_vec = nx * ws0[:, :, :, None]

        # wake deltas are in wake frame, rotate back to global frame:
        ny = np.stack((-nx[..., 1], nx[..., 0]), axis=-1)
        delta_uv = (
            wake_deltas["U"][:, :, :, None] * nx + wake_deltas["V"][:, :, :, None] * ny
        )
        del ws0, nx, ny

        # add ambient result to wake deltas:
        wind_vec += delta_uv
        del delta_uv

        # deduce WS and WD deltas:
        new_wd = uv2wd(wind_vec)
        new_ws = np.linalg.norm(wind_vec, axis=-1)
        wake_deltas[FV.WS] += new_ws - amb_results[FV.WS]
        wake_deltas[FV.WD] += new_wd - amb_results[FV.WD]



import numpy as np
import pandas as pd

from foxes.core import TurbineModel
from foxes.utils import PandasFileHelper
import foxes.variables as FV
import foxes.constants as FC


class SectorManagement(TurbineModel):
    """
    Changes variables based on variable range conditions.

    Attributes
    ----------
    source: str or pandas.DataFrame
        The file path or data

    :group: models.turbine_models

    """

    def __init__(
        self,
        data_source,
        range_vars,
        target_vars,
        col_tinds=None,
        col_tnames=None,
        colmap={},
        var_periods={FV.WD: 360.0, FV.AMB_WD: 360.0},
        pd_file_read_pars={},
    ):
        """
        Constructor.

        Parameters
        ----------
        data_source: str or pandas.DataFrame
            The file path or data
        range_vars: list of str
            The variables for which (min, max) ranges
            are specified in the data
        target_vars: list of str
            The variables that change if range variables
            are within specified ranges
        col_tinds: str, optional
            The turbine index column name in the data
        col_tnames: str, optional
            The turbine name column name in the data
        colmap: dict
            Mapping from expected to existing
            column names
        var_periods: dict
            Periods for periodic variables
        pd_file_read_pars: dict
            Parameters for pandas file reading

        """
        super().__init__()

        self.source = data_source

        self._col_i = col_tinds
        self._col_t = col_tnames
        self._rvars = range_vars
        self._tvars = target_vars
        self._colmap = colmap
        self._perds = var_periods
        self._rpars = pd_file_read_pars

        self._rdata = None
        self._tdata = None
        self._trbs = None

    def initialize(self, algo, verbosity=0):
        """
        Initializes the model.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        verbosity: int
            The verbosity level, 0 = silent

        """
        super().initialize(algo, verbosity)

        if isinstance(self.source, pd.DataFrame):
            data = self.source
        else:
            if verbosity > 0:
                print(f"{self.name}: Reading file {self.source}")
            data = PandasFileHelper.read_file(self.source, **self._rpars)

        if self._trbs is None:
            if self._col_i is not None and self._col_t is None:
                data.reset_index(inplace=True)
            elif self._col_i is None and self._col_t is not None:
                tnames = algo.farm.turbine_names
                inds = [tnames.index(name) for name in data[self._col_t]]
                data[FC.TURBINE] = inds
                self._col_i = FC.TURBINE
            else:
                raise KeyError(
                    f"{self.name}: Please either specify 'col_tinds' or 'col_tnames'"
                )
            self._trbs = data[self._col_i].to_numpy()
        n_trbs = len(self._trbs)

        self._rcols = []
        for v in self._rvars:
            col_vmin = f"{v}_min"
            col_vmin = self._colmap.get(col_vmin, col_vmin)
            if col_vmin not in data.columns:
                raise KeyError(
                    f"{self.name}: Missing column '{col_vmin}', maybe add it to 'colmap'?"
                )

            col_vmax = f"{v}_max"
            col_vmax = self._colmap.get(col_vmax, col_vmax)
            if col_vmax not in data.columns:
                raise KeyError(
                    f"{self.name}: Missing column '{col_vmax}', maybe add it to 'colmap'?"
                )

            self._rcols += [col_vmin, col_vmax]

        self._tcols = []
        for v in self._tvars:
            col = self._colmap.get(v, v)
            if col not in data.columns:
                raise KeyError(
                    f"{self.name}: Missing column '{col}', maybe add it to 'colmap'?"
                )
            self._tcols.append(col)

        n_rvars = len(self._rvars)
        self._rdata = data[self._rcols].to_numpy().reshape(n_trbs, n_rvars, 2)
        self._tdata = data[self._tcols].to_numpy()

        for vi, v in enumerate(self._rvars):
            if v in self._perds:
                self._rdata[:, vi] = np.mod(self._rdata[:, vi], self._perds[v])

    def output_farm_vars(self, algo):
        """
        The variables which are being modified by the model.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm

        Returns
        -------
        output_vars: list of str
            The output variable names

        """
        return self._tvars

    def calculate(self, algo, mdata, fdata, st_sel):
        """
        The main model calculation.

        This function is executed on a single chunk of data,
        all computations should be based on numpy arrays.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        mdata: foxes.core.MData
            The model data
        fdata: foxes.core.FData
            The farm data
        st_sel: slice or numpy.ndarray of bool
            The state-turbine selection,
            for shape: (n_states, n_turbines)

        Returns
        -------
        results: dict
            The resulting data, keys: output variable str.
            Values: numpy.ndarray with shape (n_states, n_turbines)

        """
        # prepare:
        n_trbs = len(self._trbs)
        if n_trbs == fdata.n_turbines and np.all(
            self._trbs == np.arange(fdata.n_turbines)
        ):
            tsel = np.s_[:]
        else:
            tsel = self._trbs

        # find state-turbine data that matches ranges:
        rsel = np.ones((fdata.n_states, n_trbs), dtype=bool)
        for vi, v in enumerate(self._rvars):
            d = fdata[v][:, tsel]
            if v in self._perds:
                d = np.mod(d, self._perds[v])
                mi = self._rdata[:, vi, 0]
                ma = self._rdata[:, vi, 1]
                sel = ma < mi
                if np.any(sel):
                    rsel[:, sel] = rsel[:, sel] & (
                        (d[:, sel] >= mi[sel]) | (d[:, sel] < ma[sel])
                    )
                if np.any(~sel):
                    rsel[:, ~sel] = (
                        rsel[:, ~sel]
                        & (d[:, ~sel] >= mi[~sel])
                        & (d[:, ~sel] < ma[~sel])
                    )
            else:
                rsel = (
                    rsel
                    & (d >= self._rdata[None, :, vi, 0])
                    & (d < self._rdata[None, :, vi, 1])
                )

        # set target data:
        if np.any(rsel):
            sel = np.where(rsel)
            selt = self._trbs[sel[1]]
            for vi, v in enumerate(self._tvars):
                fdata[v][sel[0], selt] = self._tdata[None, sel[1], vi]

        return {v: fdata[v] for v in self._tvars}



class Dict(dict):
    """
    A slightly enhanced dictionary.

    Attributes
    ----------
    name: str
        The dictionary name

    :group: utils

    """

    def __init__(self, *args, name=None, **kwargs):
        """
        Constructor.

        Parameters
        ----------
        *args: tuple, optional
            Arguments passed to `dict`
        name: str, optional
            The dictionary name
        **kwargs: dict, optional
            Arguments passed to `dict`

        """
        super().__init__(*args, **kwargs)
        self.name = name if name is not None else type(self).__name__

    def __getitem__(self, key):
        try:
            return super().__getitem__(key)
        except KeyError:
            k = ", ".join(sorted(list(self.keys())))
            e = f"{self.name}: Cannot find key '{key}'. Known keys: {k}"
            raise KeyError(e)



import numpy as np

from foxes.core import TurbineInductionModel
from foxes.utils import uv2wd, wd2uv
import foxes.variables as FV
import foxes.constants as FC


class VortexSheet(TurbineInductionModel):
    """
    The Vortex Sheet model implemented with a radial dependency

    Ref: Medici, D., et al. "The upstream flow of a wind turbine: blockage effect." Wind Energy 14.5 (2011): 691-697.
    https://doi.org/10.1002/we.451

    Attributes
    ----------
    pre_rotor_only: bool
        Calculate only the pre-rotor region
    induction: foxes.core.AxialInductionModel or str
        The induction model

    :group: models.wake_models.induction

    """

    def __init__(self, pre_rotor_only=False, induction="Madsen"):
        """
        Constructor.

        Parameters
        ----------
        pre_rotor_only: bool
            Calculate only the pre-rotor region
        induction: foxes.core.AxialInductionModel or str
            The induction model

        :group: models.wake_models.induction

        """
        super().__init__()
        self.induction = induction
        self.pre_rotor_only = pre_rotor_only

    def __repr__(self):
        iname = (
            self.induction if isinstance(self.induction, str) else self.induction.name
        )
        return f"{type(self).__name__}, induction={iname})"

    def sub_models(self):
        """
        List of all sub-models

        Returns
        -------
        smdls: list of foxes.core.Model
            All sub models

        """
        return [self.induction]

    def initialize(self, algo, verbosity=0, force=False):
        """
        Initializes the model.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        verbosity: int
            The verbosity level, 0 = silent
        force: bool
            Overwrite existing data

        """
        if isinstance(self.induction, str):
            self.induction = algo.mbook.axial_induction[self.induction]
        super().initialize(algo, verbosity, force)

    def new_wake_deltas(self, algo, mdata, fdata, tdata):
        """
        Initialize wake delta storage.

        They are added on the fly to the wake_deltas dict.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        mdata: foxes.core.Data
            The model data
        fdata: foxes.core.Data
            The farm data
        pdata: foxes.core.Data
            The evaluation point data
        wake_deltas: dict
            The wake deltas storage, add wake deltas
            on the fly. Keys: Variable name str, for which the
            wake delta applies, values: numpy.ndarray with
            shape (n_states, n_points, ...)

        """
        return {FV.WS: np.zeros_like(tdata[FC.TARGETS][..., 0])}

    def contribute(
        self,
        algo,
        mdata,
        fdata,
        tdata,
        downwind_index,
        wake_coos,
        wake_deltas,
    ):
        """
        Modifies wake deltas at target points by
        contributions from the specified wake source turbines.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        mdata: foxes.core.Data
            The model data
        fdata: foxes.core.Data
            The farm data
        tdata: foxes.core.Data
            The target point data
        downwind_index: int
            The index of the wake causing turbine
            in the downwnd order
        wake_coos: numpy.ndarray
            The wake frame coordinates of the evaluation
            points, shape: (n_states, n_targets, n_tpoints, 3)
        wake_deltas: dict
            The wake deltas. Key: variable name,
            value: numpy.ndarray with shape
            (n_states, n_targets, n_tpoints, ...)

        """

        # get x, y and z. Rounding for safe x < 0 condition
        x = np.round(wake_coos[..., 0], 12)
        y = wake_coos[..., 1]
        z = wake_coos[..., 2]
        r = np.sqrt(y**2 + z**2)
        r_sph = np.sqrt(r**2 + x**2)

        # get ct
        ct = self.get_data(
            FV.CT,
            FC.STATE_TARGET_TPOINT,
            lookup="w",
            algo=algo,
            fdata=fdata,
            tdata=tdata,
            upcast=True,
            downwind_index=downwind_index,
        )

        # get ws
        ws = self.get_data(
            FV.REWS,
            FC.STATE_TARGET_TPOINT,
            lookup="w",
            algo=algo,
            fdata=fdata,
            tdata=tdata,
            upcast=True,
            downwind_index=downwind_index,
        )

        # get D
        D = self.get_data(
            FV.D,
            FC.STATE_TARGET_TPOINT,
            lookup="w",
            algo=algo,
            fdata=fdata,
            tdata=tdata,
            upcast=True,
            downwind_index=downwind_index,
        )

        sp_sel = (ct > 0) & (x <= 0)
        ws_sel = ws[sp_sel]
        ct_sel = ct[sp_sel]
        r_sph_sel = r_sph[sp_sel]
        D_sel = D[sp_sel]

        if np.any(sp_sel):
            blockage = (
                ws_sel
                * (
                    1
                    - self.induction.ct2a(ct_sel)
                    * ((1 + 2 * r_sph_sel / D_sel) * (1 + (2 * r_sph_sel / D_sel) ** 2))
                    ** (-0.5)
                )
            ) - ws_sel
            wake_deltas[FV.WS][sp_sel] += blockage

        if not self.pre_rotor_only:
            sp_sel = (
                (ct > 0) & (x > 0) & (r > D / 2)
            )  # mirror in rotor plane and inverse blockage, but not directly behind rotor
            ws_sel = ws[sp_sel]
            ct_sel = ct[sp_sel]
            r_sph_sel = r_sph[sp_sel]
            D_sel = D[sp_sel]
            if np.any(sp_sel):
                blockage = (
                    ws_sel
                    * (
                        1
                        - self.induction.ct2a(ct_sel)
                        * (
                            (1 + 2 * r_sph_sel / D_sel)
                            * (1 + (2 * r_sph_sel / D_sel) ** 2)
                        )
                        ** (-0.5)
                    )
                ) - ws_sel
                wake_deltas[FV.WS][sp_sel] += -blockage

        return wake_deltas



import numpy as np
from abc import abstractmethod

from foxes.models.wake_models.axisymmetric import AxisymmetricWakeModel
import foxes.variables as FV
import foxes.constants as FC


class TopHatWakeModel(AxisymmetricWakeModel):
    """
    Abstract base class for top-hat wake models.

    Parameters
    ----------
    induction: foxes.core.AxialInductionModel or str
        The induction model

    :group: models.wake_models

    """

    def __init__(self, superpositions, induction="Betz"):
        """
        Constructor.

        Parameters
        ----------
        superpositions: dict
            The superpositions. Key: variable name str,
            value: The wake superposition model name,
            will be looked up in model book
        induction: foxes.core.AxialInductionModel or str
            The induction model

        """
        super().__init__(superpositions)
        self.induction = induction

    def sub_models(self):
        """
        List of all sub-models

        Returns
        -------
        smdls: list of foxes.core.Model
            All sub models

        """
        return [self.induction]

    def initialize(self, algo, verbosity=0, force=False):
        """
        Initializes the model.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        verbosity: int
            The verbosity level, 0 = silent
        force: bool
            Overwrite existing data

        """
        if isinstance(self.induction, str):
            self.induction = algo.mbook.axial_induction[self.induction]
        super().initialize(algo, verbosity, force)

    @abstractmethod
    def calc_wake_radius(
        self,
        algo,
        mdata,
        fdata,
        tdata,
        downwind_index,
        x,
        ct,
    ):
        """
        Calculate the wake radius, depending on x only (not r).

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        mdata: foxes.core.MData
            The model data
        fdata: foxes.core.FData
            The farm data
        tdata: foxes.core.TData
            The target point data
        downwind_index: int
            The index in the downwind order
        x: numpy.ndarray
            The x values, shape: (n_states, n_targets)
        ct: numpy.ndarray
            The ct values of the wake-causing turbines,
            shape: (n_states, n_targets)

        Returns
        -------
        wake_r: numpy.ndarray
            The wake radii, shape: (n_states, n_targets)

        """
        pass

    @abstractmethod
    def calc_centreline(
        self,
        algo,
        mdata,
        fdata,
        tdata,
        downwind_index,
        st_sel,
        x,
        wake_r,
        ct,
    ):
        """
        Calculate centre line results of wake deltas.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        mdata: foxes.core.MData
            The model data
        fdata: foxes.core.FData
            The farm data
        tdata: foxes.core.TData
            The target point data
        downwind_index: int
            The index in the downwind order
        st_sel: numpy.ndarray of bool
            The state-target selection, for which the wake
            is non-zero, shape: (n_states, n_targets)
        x: numpy.ndarray
            The x values, shape: (n_st_sel,)
        wake_r: numpy.ndarray
            The wake radii, shape: (n_st_sel,)
        ct: numpy.ndarray
            The ct values of the wake-causing turbines,
            shape: (n_st_sel,)

        Returns
        -------
        cl_del: dict
            The centre line wake deltas. Key: variable name str,
            varlue: numpy.ndarray, shape: (n_st_sel,)

        """
        pass

    def calc_wakes_x_r(
        self,
        algo,
        mdata,
        fdata,
        tdata,
        downwind_index,
        x,
        r,
    ):
        """
        Calculate wake deltas.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        mdata: foxes.core.MData
            The model data
        fdata: foxes.core.FData
            The farm data
        tdata: foxes.core.TData
            The target point data
        downwind_index: int
            The index in the downwind order
        x: numpy.ndarray
            The x values, shape: (n_states, n_targets)
        r: numpy.ndarray
            The radial values for each x value, shape:
            (n_states, n_targets, n_yz_per_target)

        Returns
        -------
        wdeltas: dict
            The wake deltas. Key: variable name str,
            value: numpy.ndarray, shape: (n_st_sel, n_r_per_x)
        st_sel: numpy.ndarray of bool
            The state-target selection, for which the wake
            is non-zero, shape: (n_states, n_targets)

        """
        ct = self.get_data(
            FV.CT,
            FC.STATE_TARGET,
            lookup="w",
            fdata=fdata,
            tdata=tdata,
            downwind_index=downwind_index,
            algo=algo,
            upcast=True,
        )

        wake_r = self.calc_wake_radius(algo, mdata, fdata, tdata, downwind_index, x, ct)

        wdeltas = {}
        st_sel = (ct > 0) & (x > 0) & np.any(r < wake_r[:, :, None], axis=2)
        if np.any(st_sel):
            x = x[st_sel]
            r = r[st_sel]
            ct = ct[st_sel]
            wake_r = wake_r[st_sel]

            cl_del = self.calc_centreline(
                algo, mdata, fdata, tdata, downwind_index, st_sel, x, wake_r, ct
            )

            isin = r < wake_r[:, None]
            for v, wdel in cl_del.items():
                wdeltas[v] = np.where(isin, wdel[:, None], 0.0)

        return wdeltas, st_sel



import pandas as pd

from foxes.core.point_data_model import PointDataModel
from foxes.utils import PandasFileHelper
import foxes.constants as FC
import foxes.variables as FV


class SetUniformData(PointDataModel):
    """
    Set uniform data (can be state dependent)

    Attributes
    ----------
    data_source: str or pandas.DataFrame or dict
        Either a file name, or a data frame, both assuming
        state dependent data. Or a dict for state independent
        uniform data (i.e., scalars)
    ovars: list of str
        The variables to be written
    var2col: dict
        Mapping from variable names to data column names

    :group: models.point_models

    """

    def __init__(
        self,
        data_source,
        output_vars,
        var2col={},
        pd_read_pars={},
    ):
        """
        Constructor.

        Parameters
        ----------
        data_source: str or pandas.DataFrame or dict
            Either a file name, or a data frame, both assuming
            state dependent data. Or a dict for state independent
            uniform data (i.e., scalars)
        output_vars: list of str
            The variables to be written
        var2col: dict
            Mapping from variable names to data column names
        pd_read_pars: dict
            pandas file reading parameters

        """
        self.data_source = data_source
        self.ovars = output_vars
        self.var2col = var2col

        self._rpars = pd_read_pars

    def load_data(self, algo, verbosity=0):
        """
        Load and/or create all model data that is subject to chunking.

        Such data should not be stored under self, for memory reasons. The
        data returned here will automatically be chunked and then provided
        as part of the mdata object during calculations.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        verbosity: int
            The verbosity level, 0 = silent

        Returns
        -------
        idata: dict
            The dict has exactly two entries: `data_vars`,
            a dict with entries `name_str -> (dim_tuple, data_ndarray)`;
            and `coords`, a dict with entries `dim_name_str -> dim_array`

        """
        self.VARS = self.var("VARS")
        self.DATA = self.var("DATA")

        if isinstance(self.data_source, pd.DataFrame):
            data = self.data_source[
                [self.var2col.get(v, v) for v in self.ovars]
            ].to_numpy(FC.DTYPE)
        elif isinstance(self.data_source, dict):
            pass
        else:
            if verbosity:
                print(f"States '{self.name}': Reading file {self.data_source}")
            rpars = dict(index_col=0)
            rpars.update(self._rpars)
            data = PandasFileHelper().read_file(self.data_source, **rpars)
            data = data[[self.var2col.get(v, v) for v in self.ovars]].to_numpy(FC.DTYPE)

        idata = super().load_data(algo, verbosity)
        idata["coords"][self.VARS] = self.ovars
        idata["data_vars"][self.DATA] = ((FC.STATE, self.VARS), data)

        return idata

    def output_point_vars(self, algo):
        """
        The variables which are being modified by the model.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm

        Returns
        -------
        output_vars: list of str
            The output variable names

        """
        return self.ovars

    def calculate(self, algo, mdata, fdata, pdata):
        """ "
        The main model calculation.

        This function is executed on a single chunk of data,
        all computations should be based on numpy arrays.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        mdata: foxes.core.MData
            The model data
        fdata: foxes.core.FData
            The farm data
        tdata: foxes.core.TData
            The target point data

        Returns
        -------
        results: dict
            The resulting data, keys: output variable str.
            Values: numpy.ndarray with shape (n_states, n_points)

        """
        for v in self.ovars:
            if self.DATA in mdata:
                pdata[v][:] = mdata[v][None, self.ovars.index(v)]
            else:
                pdata[v][:] = self.data_source[v]

        return {v: pdata[v] for v in self.ovars}



from abc import abstractmethod

from foxes.utils import all_subclasses

from .model import Model


class PartialWakesModel(Model):
    """
    Abstract base class for partial wakes models.

    Partial wakes models compute wake effects
    for rotor effective quantities.

    Attributes
    ----------
    wake_models: list of foxes.core.WakeModel
        The wake model selection
    wake_frame: foxes.core.WakeFrame, optional
        The wake frame

    :group: core

    """

    def check_wmodel(self, wmodel, error=True):
        """
        Checks the wake model type

        Parameters
        ----------
        wmodel: foxes.core.WakeModel
            The wake model to be tested
        error: bool
            Flag for raising TypeError

        Returns
        -------
        chk: bool
            True if wake model is compatible

        """
        return True

    @abstractmethod
    def get_wake_points(self, algo, mdata, fdata):
        """
        Get the wake calculation points, and their
        weights.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        mdata: foxes.core.Data
            The model data
        fdata: foxes.core.Data
            The farm data

        Returns
        -------
        rpoints: numpy.ndarray
            The wake calculation points, shape:
            (n_states, n_turbines, n_tpoints, 3)
        rweights: numpy.ndarray
            The target point weights, shape: (n_tpoints,)

        """
        pass

    def new_wake_deltas(self, algo, mdata, fdata, tdata, wmodel):
        """
        Creates new initial wake deltas, filled
        with zeros.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        mdata: foxes.core.Data
            The model data
        fdata: foxes.core.Data
            The farm data
        tdata: foxes.core.Data
            The target point data
        wmodel: foxes.core.WakeModel
            The wake model

        Returns
        -------
        wake_deltas: dict
            Key: variable name, value: The zero filled
            wake deltas, shape: (n_states, n_turbines, n_tpoints, ...)

        """
        return wmodel.new_wake_deltas(algo, mdata, fdata, tdata)

    def contribute(
        self,
        algo,
        mdata,
        fdata,
        tdata,
        downwind_index,
        wake_deltas,
        wmodel,
    ):
        """
        Modifies wake deltas at target points by
        contributions from the specified wake source turbines.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        mdata: foxes.core.MData
            The model data
        fdata: foxes.core.FData
            The farm data
        tdata: foxes.core.TData
            The target point data
        downwind_index: int
            The index of the wake causing turbine
            in the downwnd order
        wake_deltas: dict
            The wake deltas. Key: variable name,
            value: numpy.ndarray with shape
            (n_states, n_targets, n_tpoints, ...)
        wmodel: foxes.core.WakeModel
            The wake model

        """
        wcoos = algo.wake_frame.get_wake_coos(algo, mdata, fdata, tdata, downwind_index)
        wmodel.contribute(algo, mdata, fdata, tdata, downwind_index, wcoos, wake_deltas)

    @abstractmethod
    def finalize_wakes(
        self,
        algo,
        mdata,
        fdata,
        tdata,
        amb_res,
        rpoint_weights,
        wake_deltas,
        wmodel,
        downwind_index,
    ):
        """
        Updates the wake_deltas at the selected target
        downwind index.

        Modifies wake_deltas on the fly.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        mdata: foxes.core.MData
            The model data
        fdata: foxes.core.FData
            The farm data
        tdata: foxes.core.Data
            The target point data
        amb_res: dict
            The ambient results at the target points
            of all rotors. Key: variable name, value
            np.ndarray of shape:
            (n_states, n_turbines, n_rotor_points)
        rpoint_weights: numpy.ndarray
            The rotor point weights, shape: (n_rotor_points,)
        wake_deltas: dict
            The wake deltas. Key: variable name,
            value: np.ndarray of shape
            (n_states, n_turbines, n_tpoints)
        wmodel: foxes.core.WakeModel
            The wake model
        downwind_index: int
            The index in the downwind order

        Returns
        -------
        final_wake_deltas: dict
            The final wake deltas at the selected downwind
            turbines. Key: variable name, value: np.ndarray
            of shape (n_states, n_rotor_points)

        """
        pass

    @classmethod
    def new(cls, pwake_type, **kwargs):
        """
        Run-time partial wakes factory.

        Parameters
        ----------
        pwake_type: str
            The selected derived class name

        """

        if pwake_type is None:
            return None

        allc = all_subclasses(cls)
        found = pwake_type in [scls.__name__ for scls in allc]

        if found:
            for scls in allc:
                if scls.__name__ == pwake_type:
                    return scls(**kwargs)

        else:
            estr = "Partial wakes model type '{}' is not defined, available types are \n {}".format(
                pwake_type, sorted([i.__name__ for i in allc])
            )
            raise KeyError(estr)
